#include <common.h>
#include <jffs2/jffs2.h>
#include <asm/addrspace.h>
#include <asm/types.h>
#include "ar7100_soc.h"

#ifdef	CONFIG_STATUS_LED
#include <status_led.h>
#endif	//CONFIG_STATUS_LED

#ifndef	CONFIG_BUFFALO		//for Design
#error	"Need BUFFALO configuration"
#endif	//CONFIG_BUFFALO	//for Design

//----------------------------------------------------------------------------
// Common device details.
#define FLASHWORD(_x)			((CFG_FLASH_WORD_SIZE)(_x))

#define FLASH_Read_ID                   FLASHWORD( 0x90 )
#define FLASH_Read_ID_Exit              FLASHWORD( 0xF0 )
#define FLASH_Reset                     FLASHWORD( 0xFF )
#define FLASH_Program                   FLASHWORD( 0xA0 )
#define FLASH_Block_Erase               FLASHWORD( 0x30 )

#define FLASH_Data                      FLASHWORD( 0x80 )	// Data complement
#define FLASH_Busy                      FLASHWORD( 0x44 )	// "Toggle" bit
#define FLASH_Err                       FLASHWORD( 0x20 )
#define FLASH_Sector_Erase_Timer        FLASHWORD( 0x08 )

#define FLASH_Setup_Code1               FLASHWORD( 0xAA )
#define FLASH_Setup_Code2               FLASHWORD( 0x55 )
#define FLASH_Setup_Erase               FLASHWORD( 0x80 )
#define AMD_SPANSION			0x227e
#define VENDOR_ID(x)			(ushort) ((x & 0x00ff0000) >> 16)
#define DEVICE_ID(x)			(ushort) (x & 0x0000ffff)

/*
 * globals
 */
flash_info_t flash_info[CFG_MAX_FLASH_BANKS];
#ifdef	CONFIG_BUFFALO		//for Design
static ushort read_dev_id2(ushort *pid3);
#endif	//CONFIG_BUFFALO	//for Design

typedef struct ar9100_flash_geom {
    ushort vendor_id;
    ushort device_id;
#ifdef	CONFIG_BUFFALO
    uint16_t device_id2;
    uint16_t device_id3;
#endif	//CONFIG_BUFFALO
    char *name;
    ulong nsectors;
    ulong sector_size;
    ulong size;
} ar9100_flash_geom_t;

ar9100_flash_geom_t flash_geom_tbl[] = {
#ifndef	CONFIG_BUFFALO
    {0x00bf, 0x2780, "SST-39VF400", 8, 0x01000, 0x080000},	/* 512 KB */
    {0x00bf, 0x2782, "SST-39VF160", 32, 0x01000, 0x200000},	/* 2 MB   */
    {0x00bf, 0x236b, "SST-39VF6401", 128, 0x01000, 0x800000},	/* 8 MB   */
    {0x00bf, 0x236a, "SST-39VF6402", 128, 0x01000, 0x800000},	/* 8 MB   */
    {0x00bf, 0x236d, "SST-39VF6402", 128, 0x01000, 0x800000},	/* 8 MB   */
    {0x0001, 0x227e, "AMD-SPANSION", 128, 0x02000, 0x1000000},	/* 16 MB  */
    {0xffff, 0xffff, NULL, 0, 0}	/* end list */
#else	//CONFIG_BUFFALO
    {0x0000, 0x0000, 0x0000, 0x0000, "UNKNOWN",			  0, 0x00000, 0x0000000},  /* 0 MB  */
    //	Spansion
    {0x0001, 0x227e, 0x2223, 0x2201, "S29GL512N",		512, 0x02000, 0x4000000},  /* 64 MB  */
    {0x0001, 0x227e, 0x2222, 0x2201, "S29GL256N",		256, 0x02000, 0x2000000},  /* 32 MB  */
    {0x0001, 0x227e, 0x2221, 0x2201, "S29GL128N",		128, 0x02000, 0x1000000},  /* 16 MB  */
    {0x0001, 0x227e, 0x2212, 0x2201, "S29GL256M",		256, 0x02000, 0x2000000},  /* 32 MB  */
    {0x0001, 0x227e, 0x2212, 0x2200, "S29GL128M",		128, 0x02000, 0x1000000},  /* 16 MB  */

    {0x0001, 0x227e, 0x2213, 0x2200, "S29GL064M-R0",	128, 0x01000, 0x0800000},  /*  8 MB  */
    {0x0001, 0x227e, 0x220C, 0x2201, "S29GL064M-R1",	128, 0x01000, 0x0800000},  /*  8 MB  */
    {0x0001, 0x227e, 0x2210, 0x2201, "S29GL064M-R3",	128, 0x01000, 0x0800000},  /*  8 MB  */
    {0x0001, 0x227e, 0x2210, 0x2200, "S29GL064M-R4",	128, 0x01000, 0x0800000},  /*  8 MB  */
    {0x0001, 0x227e, 0x2213, 0x2201, "S29GL064M-R7",	128, 0x01000, 0x0800000},  /*  8 MB  */

    {0x0001, 0x227e, 0x221c, 0x2200, "S29GL032M-R0",	 64, 0x01000, 0x0400000},  /*  4 MB  */
    {0x0001, 0x227e, 0x221d, 0x2200, "S29GL032M-R1",	 64, 0x01000, 0x0400000},  /*  4 MB  */
    {0x0001, 0x227e, 0x221a, 0x2201, "S29GL032M-R3",	 64, 0x01000, 0x0400000},  /*  4 MB  */
    {0x0001, 0x227e, 0x221a, 0x2200, "S29GL032M-R4",	 64, 0x01000, 0x0400000},  /*  4 MB  */
    {0x0001, 0x227e, 0x221c, 0x2201, "S29GL032M-R7",	 64, 0x01000, 0x0400000},  /*  4 MB  */

    //	Numonyx
    {0x0089, 0x227e, 0x2222, 0x2201, "M29EWH-256M",		256, 0x02000, 0x2000000},  /* 32 MB  */
    {0x0089, 0x227e, 0x2222, 0x2200, "M29EWL-256M",		256, 0x02000, 0x2000000},  /* 32 MB  */

    //	Macronix(MX)
    {0x00C2, 0x227e, 0x2222, 0x2201, "MX29GL256E",		256, 0x02000, 0x2000000},  /* 32 MB  */

    //	end of list
    {0xffff, 0xffff, 0xffff, 0xffff, NULL, 0, 0, 0}	/* end list */
#endif	//CONFIG_BUFFALO
};


/*
 * statics
 */
static int write_word(flash_info_t * info, ulong dest, ulong data);
static ulong read_id(void);

unsigned long flash_init(void)
{
    ulong sector_size;
    ar9100_flash_geom_t *geom;
    ushort venid, devid;
#ifdef	CONFIG_BUFFALO		//for Design
	ushort devid2;
	ushort devid3;
#endif	//CONFIG_BUFFALO	//for Design

    int i;

    flash_info->flash_id = read_id();
#ifdef	CONFIG_BUFFALO		//for Design
    devid2 = read_dev_id2(&devid3);
#endif	//CONFIG_BUFFALO	//for Design
    venid = VENDOR_ID(flash_info->flash_id);
    devid = DEVICE_ID(flash_info->flash_id);

    for (i = 0; flash_geom_tbl[i].name != NULL; i++) {
#ifndef	CONFIG_BUFFALO
        if (venid == flash_geom_tbl[i].vendor_id &&
            devid == flash_geom_tbl[i].device_id) {
            break;
        }
#else	//CONFIG_BUFFALO
        if ((venid == flash_geom_tbl[i].vendor_id) &&
            (devid == flash_geom_tbl[i].device_id) &&
            (devid2== flash_geom_tbl[i].device_id2) &&
            (devid3== flash_geom_tbl[i].device_id3)
         ) {
            break;
        }
#endif	//CONFIG_BUFFALO
    }

    geom = &flash_geom_tbl[i];
    if (geom->name == NULL) {
        printf("Unknown flash device\n");
#ifndef	CONFIG_BUFFALO
        return -1;
#else	//CONFIG_BUFFALO
        i	= 0;
        geom	= &flash_geom_tbl[i];
#endif	//CONFIG_BUFFALO
    }

#ifndef	CONFIG_BUFFALO		//for Design
#else	//CONFIG_BUFFALO	//for Design
	if (i==0) {
		geom->size = 1ul << (*(volatile CFG_FLASH_WORD_SIZE *)(CFG_FLASH_BASE + 0x4e));
		geom->nsectors = ((*(volatile CFG_FLASH_WORD_SIZE *)(CFG_FLASH_BASE + 0x5a)) & 0xFF)
								 | ((*(volatile CFG_FLASH_WORD_SIZE *)(CFG_FLASH_BASE + 0x5c)) << 8);
		geom->nsectors++ ;
		geom->sector_size	= geom->size / geom->nsectors;
   }
#endif	//CONFIG_BUFFALO	//for Design
    flash_info->size = geom->size;	/* bytes */
    flash_info->sector_count = geom->nsectors;
    sector_size = (geom->sector_size * 16); /* One Erase block size */

    for (i = 0; i < flash_info->sector_count; i++) {
        flash_info->start[i] = CFG_FLASH_BASE + (i * sector_size);
        flash_info->protect[i] = 0;
    }
    flash_print_info(flash_info);

    return (flash_info->size);
}


void flash_print_info(flash_info_t * info)
{
    printf("Name: ");
    switch (DEVICE_ID(info->flash_id))
    {
    case 0x2780: printf ("SST-39VF400");            break;
    case 0x2782: printf ("SST-39VF160");            break;
    case 0x236b: printf ("SST-39VF6401");           break;
    case 0x236a: printf ("SST-39VF6402");           break;
    case 0x236d: printf ("SST-39VF6402");           break;
    case 0x227e: printf ("AMD-SPANSION");           break;
    default:     printf ("Unknown Vendor ");        break;
    }

    printf(" Flash id: 0x%X, Size: %lu bytes.\n", info->flash_id, info->size);
}

int flash_erase(flash_info_t * info, int s_first, int s_last)
{
    int i = s_first;
    volatile CFG_FLASH_WORD_SIZE *ROM =
        (volatile CFG_FLASH_WORD_SIZE *) (info->start[0]);
    int timeout;
#ifdef	CONFIG_STATUS_LED
    int	led_cnt	= 0;
#endif	//CONFIG_STATUS_LED

    printf("First %#x last %#x\n", s_first, s_last);

#ifdef	CONFIG_STATUS_LED
	status_led_set(STATUS_LED_DIAG, STATUS_LED_ON);
#endif	//CONFIG_STATUS_LED

    for (i = s_first; i <= s_last; i++) {
        CFG_FLASH_WORD_SIZE state, prev_state, rd_data;
        volatile CFG_FLASH_WORD_SIZE *addr_ptr =
            (volatile CFG_FLASH_WORD_SIZE *) (info->start[i]);

        // Program data [byte] - 6 step sequence
        ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
        ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
        ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Erase;
        ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
        ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
        *addr_ptr = FLASH_Block_Erase;

        printf("\b\b\b\b\b %3d\%",(( ((i-s_first) + 1) * 100) / ((s_last-s_first) + 1 )) );

#if 1
            // Wait for erase completion.
            timeout = 10000000;
            while (timeout) {
		state = *addr_ptr;
		if (FLASHWORD(0xffff) == state) {
	    		break;
		}
		timeout--;
            }
#else
            // Wait for completion (bit 6 stops toggling)
            timeout = 5000000;
            prev_state = *addr_ptr & FLASH_Busy;

            while (timeout) {
		rd_data = *addr_ptr;
		state = rd_data & FLASH_Busy;
		if ((prev_state == state) && (rd_data == FLASHWORD(0xffff))) {
	    		break;
	       }
	       timeout--;
	       prev_state = state;
           }
#endif
        if (!timeout){
	    printf("Error erasing flash...\n");
            return -1;
        }

#ifdef	CONFIG_STATUS_LED
		status_led_set(STATUS_LED_DIAG, (led_cnt++ & 0x01) ? STATUS_LED_ON : STATUS_LED_OFF);
#endif	//CONFIG_STATUS_LED
    }
        printf("\n");

#ifdef	CONFIG_STATUS_LED
	status_led_set(STATUS_LED_DIAG, STATUS_LED_OFF);
#endif	//CONFIG_STATUS_LED

    return 0;
}

/*
 * Copy memory to flash, returns:
 * Assumption: Caller has already erased the appropriate sectors.
 * 0 - OK
 * 1 - write timeout
 */
int write_buff(flash_info_t * info, uchar * src, ulong addr, ulong cnt)
{
    ulong cp, wp, data;
    int i, l, rc, j=0, count;
#ifdef	CONFIG_BUFFALO
    int percent=0;
#endif	//CONFIG_BUFFALO
#ifdef	CONFIG_STATUS_LED
    int	led_cnt	= 0;
#endif	//CONFIG_STATUS_LED

    wp = (addr & ~3);   /* get lower word aligned address */
    count=cnt;
    printf("\n");

#ifdef	CONFIG_STATUS_LED
	status_led_set(STATUS_LED_DIAG, STATUS_LED_ON);
#endif	//CONFIG_STATUS_LED
    /*
     * handle unaligned start bytes
     */
    if ((l = addr - wp) != 0) {
        data = 0;
        for (i = 0, cp = wp; i < l; ++i, ++cp) {
            data = (data << 8) | (*(uchar *) cp);
        }
        for (; i < 4 && cnt > 0; ++i) {
            data = (data << 8) | *src++;
            --cnt;
            ++cp;
        }
        for (; cnt == 0 && i < 4; ++i, ++cp) {
            data = (data << 8) | (*(uchar *) cp);
        }

        if ((rc = write_word(info, wp, data)) != 0) {
            return (rc);
        }
	j = j + 4;
        wp += 4;
    }

    /*
     * handle word aligned part
     */
    while (cnt >= 4) {
        data = 0;
        for (i = 0; i < 4; ++i) {
            data = (data << 8) | *src++;
        }
#ifndef	CONFIG_BUFFALO
        if(!(j%1024))
	   printf("\b\b\b\b\b %3d\%",((j * 100) / count));
#else	//CONFIG_BUFFALO
		if (!(j%8192)) {
			int		new_percent	= ((unsigned)j * 100u) / count;
			if (new_percent != percent) {
				percent	= new_percent;
				printf("\b\b\b\b\b %3d\%",percent);
			}
		}
#endif	//CONFIG_BUFFALO
        if ((rc = write_word(info, wp, data)) != 0) {
            return (rc);
        }
        wp += 4;
        cnt -= 4;
        j += 4;
#ifdef	CONFIG_STATUS_LED
        if (!(j & 0xFFFF))
			status_led_set(STATUS_LED_DIAG, (led_cnt++ & 0x01) ? STATUS_LED_ON : STATUS_LED_OFF);
#endif	//CONFIG_STATUS_LED
    }

#ifndef	CONFIG_BUFFALO
    printf("\b\b\b\b\b %3d\%\n",(((j + 4) * 100 ) / count));
#else	//CONFIG_BUFFALO
    printf("\b\b\b\b\b %3d\%\n",((((unsigned)j + 4) * 100u ) / count));
#endif	//CONFIG_BUFFALO
    if (cnt == 0) {
        return (0);
    }

    /*
     * handle unaligned tail bytes
     */
    data = 0;
    for (i = 0, cp = wp; i < 4 && cnt > 0; ++i, ++cp) {
        data = (data << 8) | *src++;
        --cnt;
    }
    for (; i < 4; ++i, ++cp) {
        data = (data << 8) | (*(uchar *) cp);
    }
#ifdef	CONFIG_STATUS_LED
    {
    	int	retval	= write_word(info, wp, data);
		status_led_set(STATUS_LED_DIAG, STATUS_LED_OFF);
		return	retval;
    }
#endif	//CONFIG_STATUS_LED
    return (write_word(info, wp, data));
}

static int write_word(flash_info_t * info, ulong dest, ulong data)
{
    volatile CFG_FLASH_WORD_SIZE *ROM =
        (CFG_FLASH_WORD_SIZE *) (info->start[0]);
    volatile CFG_FLASH_WORD_SIZE *dest2 = (CFG_FLASH_WORD_SIZE *) dest;
    CFG_FLASH_WORD_SIZE *data2 = (CFG_FLASH_WORD_SIZE *) & data;
    int i;

    /* Check if Flash is (sufficiently) erased */
    if ((*((vu_long *) dest) & data) != data) {
        return (2);
    }

    for (i = 0; i < 4 / sizeof(CFG_FLASH_WORD_SIZE); i++) {
        CFG_FLASH_WORD_SIZE state, prev_state;
        int timeout;

        ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
        ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
        ROM[CFG_FLASH_ADDR0] = FLASH_Program;
        dest2[i] = data2[i];

#if 1
            timeout = 10000000;
            while (timeout) {
		if (dest2[i] == data2[i]) {
	   		 break;
		}
		timeout--;
            }
#else
            // Wait for completion (bit 6 stops toggling)
            timeout = 5000000;
            prev_state = dest2[i] & FLASH_Busy;
            while (timeout) {
		state = dest2[i] & FLASH_Busy;
		if (prev_state == state) {
	    		break;
		}
		timeout--;
		prev_state = state;
            }
#endif

        if (!timeout) {
		printf("Error while Writing into flash...\n");
            return -1;
	}
    }

    return (0);
}
#ifdef	CONFIG_BUFFALO		//for Design
static ushort read_dev_id2(ushort *pid3)
{
    volatile CFG_FLASH_WORD_SIZE *ROM;
    ushort id2;
    ushort id3;

    ROM = (volatile CFG_FLASH_WORD_SIZE *) CFG_FLASH_BASE;

    ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
    ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
    ROM[CFG_FLASH_ADDR0] = FLASH_Read_ID;

    udelay(10000);

    //	dummy read
    id2	= ROM[0];
    id2	= ROM[1];
    id2 = ROM[14];
    id3 = ROM[15];
    printf("Dev ID2:%X\n",id2);
    printf("Dev ID3:%X\n",id3);
    if (pid3)	*pid3	= id3;

    ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
    ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
    ROM[CFG_FLASH_ADDR0] = FLASH_Read_ID_Exit;

    udelay(10000);
    return id2;
}
#endif	//CONFIG_BUFFALO	//for Design

static ulong read_id()
{
    volatile CFG_FLASH_WORD_SIZE *ROM;
    ulong id;

    ROM = (volatile CFG_FLASH_WORD_SIZE *) CFG_FLASH_BASE;

    ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
    ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
    ROM[CFG_FLASH_ADDR0] = FLASH_Read_ID;

    udelay(10000);

    id = ROM[0];
    id = (id << 16) | (ROM[1] & 0xffff);

    ROM[CFG_FLASH_ADDR0] = FLASH_Setup_Code1;
    ROM[CFG_FLASH_ADDR1] = FLASH_Setup_Code2;
    ROM[CFG_FLASH_ADDR0] = FLASH_Read_ID_Exit;

    udelay(10000);
    return id;
}
