#include <config.h>
#include <linux/types.h>
#include <common.h>
#include <miiphy.h>
#include <asm/addrspace.h>
#include "rtl8366rb_phy.h"
#include "rtl8368s_errno.h"

#define	DEBUG_MSG
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//
//	copy from "rtl8368s_asicdrv.c"
//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

/*=======================================================================
 * 1. Asic read/write driver through SMI
 *========================================================================*/
/*
@func int32 | rtl8368s_setAsicRegBit | Set a bit value of a specified register.
@parm uint32 | reg | Register's address.
@parm uint32 | bit | Bit location. For 16-bits register only. Maximun value is 15 for MSB location.
@parm uint32 | value | Value to set. It can be value 0 or 1.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue TBLDRV_EINVALIDINPUT | Invalid input parameter. 
@comm
	Set a bit of a specified register to 1 or 0. It is 16-bits system of RTL8366s chip.
	
*/
int32 rtl8368s_setAsicRegBit(uint32 reg, uint32 bit, uint32 value)
{

#if defined(RTK_X86_ASICDRV)
	uint32 regData;
	int32 retVal;
	
	if(bit>=RTL8368S_REGBITLENGTH)
		return ERRNO_INPUT;

	retVal = RTL_SINGLE_READ(reg, 2, &regData);
	if (retVal != TRUE) return ERRNO_SMI;

	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);

	if (value) 
		regData = regData | (1<<bit);
	else
		regData = regData & ~(1<<bit);

	retVal = RTL_SINGLE_WRITE(reg, regData);
	if (retVal != TRUE) return ERRNO_SMI;

	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,regData);

	
#elif defined(CONFIG_RTL8368S_ASICDRV_TEST)

	/*MIBs emulating*/
	if(reg == RTL8368S_MIB_CTRL_REG)
	{
		reg = RTL8368S_MIB_CTRL_VIRTUAL_ADDR;
	}	
	else if(reg >=RTL8368S_PHY_ACCESS_CTRL_REG)
	{
		reg = reg -RTL8368S_PHY_ACCESS_CTRL_REG + RTL8368S_PHY_VIRTUAL_ADDR;
	}
	else if(bit>=RTL8368S_REGBITLENGTH)
		return ERRNO_INPUT;
	else if(reg >= RTL8368S_VIRTUAL_REG_SIZE)
		return ERRNO_REGUNDEFINE;

	if (value) 
	{
		Rtl8368sVirtualReg[reg] =  Rtl8368sVirtualReg[reg] | (1<<bit);

	}
	else
	{
		Rtl8368sVirtualReg[reg] =  Rtl8368sVirtualReg[reg] & (~(1<<bit));
	}
	
	if(0x8368 == cleDebuggingDisplay)
		PRINT("--[0x%4.4x]=0x%4.4x\n",reg,Rtl8368sVirtualReg[reg]);

	
#else
	uint32 regData;
	int32 retVal;
	
	if(bit>=RTL8368S_REGBITLENGTH)
		return ERRNO_INPUT;

	retVal = smi_read(reg, &regData);
	if (retVal != SUCCESS) return ERRNO_SMI;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);
#endif

	if (value) 
		regData = regData | (1<<bit);
	else
		regData = regData & (~(1<<bit));
	
	retVal = smi_write(reg, regData);
	if (retVal != SUCCESS) return ERRNO_SMI;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,regData);
#endif

#endif
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicRegBits | Set bits value of a specified register.
@parm uint32 | reg | Register's address.
@parm uint32 | bits | Bits mask for setting. 
@parm uint32 | value | Bits value for setting. Value of bits will be set with mapping mask bit is 1.   
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue TBLDRV_EINVALIDINPUT | Invalid input parameter. 
@comm
	Set bits of a specified register to value. Both bits and value are be treated as bit-mask.
	
*/
int32 rtl8368s_setAsicRegBits(uint32 reg, uint32 bits, uint32 value)
{
	
#if defined(RTK_X86_ASICDRV)//RTK-CNSD2-NickWu-20061222: for x86 compile

	uint32 regData;
	int32 retVal;
	
	if(bits>= (1<<RTL8368S_REGBITLENGTH) )
		return ERRNO_INPUT;

	retVal = RTL_SINGLE_READ(reg, 2, &regData);
	if (retVal != TRUE) return ERRNO_SMI;
	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);

	regData = regData & (~bits);
	regData = regData | (value & bits);

	retVal = RTL_SINGLE_WRITE(reg, regData);
	if (retVal != TRUE) return ERRNO_SMI;
	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,regData);
	
#elif defined(CONFIG_RTL8368S_ASICDRV_TEST)
	uint32 regData;

	/*MIBs emulating*/
	if(reg == RTL8368S_MIB_CTRL_REG)
	{
		reg = RTL8368S_MIB_CTRL_VIRTUAL_ADDR;
	}	
	else if(reg >=RTL8368S_PHY_ACCESS_CTRL_REG)
	{
		reg = reg -RTL8368S_PHY_ACCESS_CTRL_REG + RTL8368S_PHY_VIRTUAL_ADDR;
	}
	else if(reg >= RTL8368S_VIRTUAL_REG_SIZE)
		return ERRNO_REGUNDEFINE;

	regData = Rtl8368sVirtualReg[reg] & (~bits);
	regData = regData | (value & bits);
	
	Rtl8368sVirtualReg[reg] = regData;

	if(0x8368 == cleDebuggingDisplay)
		PRINT("--[0x%4.4x]=0x%4.4x\n",reg,regData);
	
#else
	uint32 regData;
	int32 retVal;
	
	if(bits>= (1<<RTL8368S_REGBITLENGTH) )
		return ERRNO_INPUT;

	retVal = smi_read(reg, &regData);
	if (retVal != SUCCESS) return ERRNO_SMI;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);
#endif

	regData = regData & (~bits);
	regData = regData | (value & bits);

	
	retVal = smi_write(reg, regData);
	if (retVal != SUCCESS) return ERRNO_SMI;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,regData);
#endif

#endif
	return SUCCESS;
}

/*
@func int32 | rtl8368s_getAsicReg | Get content of register.
@parm uint32 | reg | Register's address.
@parm uint32* | value | Value of register.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
 	Value 0x0000 will be returned for ASIC un-mapping address.
	
*/
int32 rtl8368s_getAsicReg(uint32 reg, uint32 *value)
{
	
#if defined(RTK_X86_ASICDRV)//RTK-CNSD2-NickWu-20061222: for x86 compile

	uint32 regData;
	int32 retVal;

	retVal = RTL_SINGLE_READ(reg, 2, &regData);
	if (retVal != TRUE) return ERRNO_SMI;

	*value = regData;
	
	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);

#elif defined(CONFIG_RTL8368S_ASICDRV_TEST)
       uint16 virtualMIBs;

	/*MIBs emulating*/
	if(reg == RTL8368S_MIB_CTRL_REG)
	{
		reg = RTL8368S_MIB_CTRL_VIRTUAL_ADDR;
	}	
	else if(reg >= RTL8368S_MIB_START_ADDRESS && reg <= RTL8368S_MIB_END_ADDRESS)
	{
		virtualMIBs = reg;
		
		reg = (reg & 0x0003) + RTL8368S_MIB_READ_VIRTUAL_ADDR;

		Rtl8368sVirtualReg[reg] = virtualMIBs;
	}
	else if(reg >=RTL8368S_PHY_ACCESS_CTRL_REG)
	{
		reg = reg -RTL8368S_PHY_ACCESS_CTRL_REG + RTL8368S_PHY_VIRTUAL_ADDR;
	}
	else if(reg >= RTL8368S_VIRTUAL_REG_SIZE)
		return ERRNO_REGUNDEFINE;

	*value = Rtl8368sVirtualReg[reg];

	if(0x8368 == cleDebuggingDisplay)
		PRINT("--[0x%4.4x]=0x%4.4x\n",reg,Rtl8368sVirtualReg[reg]);
	
#else
	uint32 regData;
	int32 retVal;

	retVal = smi_read(reg, &regData);
	if (retVal != SUCCESS) return ERRNO_SMI;

	*value = regData;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("R:[0x%4.4x]=0x%4.4x\n",reg,regData);
#endif

#endif
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicReg | Set content of asic register.
@parm uint32 | reg | Register's address.
@parm uint32 | value | Value setting to register.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The value will be set to ASIC mapping address only and it is always return SUCCESS while setting un-mapping address registers.
	
*/
int32 rtl8368s_setAsicReg(uint32 reg, uint32 value)
{

#if defined(RTK_X86_ASICDRV)//RTK-CNSD2-NickWu-20061222: for x86 compile

	int32 retVal;

	retVal = RTL_SINGLE_WRITE(reg, value);
	if (retVal != TRUE) return ERRNO_SMI;

	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,value);

#elif defined(CONFIG_RTL8368S_ASICDRV_TEST)
	/*MIBs emulating*/
	if(reg == RTL8368S_MIB_CTRL_REG)
	{
		reg = RTL8368S_MIB_CTRL_VIRTUAL_ADDR;
	}	
	else if(reg >= RTL8368S_MIB_START_ADDRESS && reg <= RTL8368S_MIB_END_ADDRESS)
	{
		reg = (reg & 0x0003) + RTL8368S_MIB_READ_VIRTUAL_ADDR;
	}
	else if(reg >=RTL8368S_PHY_ACCESS_CTRL_REG)
	{
		reg = reg -RTL8368S_PHY_ACCESS_CTRL_REG + RTL8368S_PHY_VIRTUAL_ADDR;
	}
	else if(reg >= RTL8368S_VIRTUAL_REG_SIZE)
		return ERRNO_REGUNDEFINE;

	Rtl8368sVirtualReg[reg] = value;

	if(0x8368 == cleDebuggingDisplay)
		PRINT("--[0x%4.4x]=0x%4.4x\n",reg,Rtl8368sVirtualReg[reg]);

#else
	int32 retVal;

	retVal = smi_write(reg, value);
	if (retVal != SUCCESS) return ERRNO_SMI;

#ifdef CONFIG_RTL865X_CLE
	if(0x8368 == cleDebuggingDisplay)
		PRINT("W:[0x%4.4x]=0x%4.4x\n",reg,value);
#endif

#endif
	return SUCCESS;
}

/*=======================================================================
 * 2. Port property APIs
 *========================================================================*/

/*
@func int32 | rtl8368s_getAsicPortLinkState | Get a specific port's link state.
@parm enum PORTID | port | Physical Port number.
@parm enum PORTLINKSPEED  * | speed | current link speed, 
	SPD_10M: 10Mbps, 
	SPD_100M: 100Mbps, 
	SPD_1000M: 1000Mbps.
@parm enum PORTLINKDUPLEXMODE * | duplex | current duplex status,
	FULL_DUPLEX: full duplex, 
	HALF_DUPLEX: half duplex.
@parm uint32  * | link | Link status, 1: link up, 0: link down.
@parm uint32  * | rxPause | The pause frame response ability, 1: active, 0: inactive.
@parm uint32  * | txPause | The pause frame transmit ability, 1: active, 0: inactive.
@parm uint32  * | nWay | N-way function, 1: enable, 0: disable.
@rvalue SUCCESS | Success.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get a specific port's link state information including link up/down, 
	link speed, auto negotiation enabled/disabled, Tx in active/inactive and Rx in 
	active/inactive. 
*/
int32 rtl8368s_getAsicPortLinkState(enum PORTID port, enum PORTLINKSPEED *speed, enum PORTLINKDUPLEXMODE *duplex,uint32 *link, uint32 *txPause,uint32 *rxPause,uint32 *nWay)
{
	int32 retVal;
	uint32 RegData;
	
	if(port <PORT0 || port >=PORT_MAX)
		return ERRNO_INPUT;

	retVal = rtl8368s_getAsicReg(RTL8368S_PORT_LINK_STATUS_BASE + (port >>1),&RegData);
	if(SUCCESS != retVal)	
		return FAILED;


	if(port&0x1)
	{
		RegData = RegData >> 8;
	}

	*speed = (RegData & RTL8368S_PORT_STATUS_SPEED_MSK)>>RTL8368S_PORT_STATUS_SPEED_OFF;
	*duplex = (RegData & RTL8368S_PORT_STATUS_DUPLEX_MSK)>>RTL8368S_PORT_STATUS_DUPLEX_OFF;
	*link = (RegData & RTL8368S_PORT_STATUS_LINK_MSK)>>RTL8368S_PORT_STATUS_LINK_OFF;
	*txPause = (RegData & RTL8368S_PORT_STATUS_TXPAUSE_MSK)>>RTL8368S_PORT_STATUS_TXPAUSE_OFF;
	*rxPause = (RegData & RTL8368S_PORT_STATUS_RXPAUSE_MSK)>>RTL8368S_PORT_STATUS_RXPAUSE_OFF;
	*nWay = (RegData & RTL8368S_PORT_STATUS_AN_MSK)>>RTL8368S_PORT_STATUS_AN_OFF;
	
	return SUCCESS;
}

/*=======================================================================
 *  VLAN stacking APIs
 *========================================================================*/
/*
@func int32 | rtl8368s_setAsicSvlan | Configure enable S-VLAN  function.
@parm uint32 | enabled | S-VLAN function 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	 There are service provider VLAN functions defined in IEEE 802.1ad standard. System supports user defined type/length for parsing S-tag frame at uplink 
	 ports. Without S-VLAN function enable setting, S-tag frame from uplink ports will be treated as normal frame payload to forward. Only S-tag frame contained
	 supported S-VID will be accepted from uplink ports.
*/
int32 rtl8368s_setAsicSvlan(uint32 enabled)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_EN_OFF,enabled);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlan | Get S-VLAN  function configuration.
@parm uint32* | enabled | S-VLAN function 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get configuration of SVLAN function enable setting.
	
*/
int32 rtl8368s_getAsicSvlan(uint32* enabled)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_EN_MSK;

	if(regData)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;
	
}

/*
@func int32 | rtl8368s_setAsicSvlanS2Cmbrsel | Configure enable S-VLAN  function.
@parm uint32 | enabled | S-VLAN S2C member selection function 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	 There are service provider VLAN functions defined in IEEE 802.1ad standard. System supports user defined type/length for parsing S-tag frame at uplink 
	 ports. Without S-VLAN function enable setting, S-tag frame from uplink ports will be treated as normal frame payload to forward. Only S-tag frame contained
	 supported S-VID will be accepted from uplink ports.
*/
int32 rtl8368s_setAsicSvlanS2Cmbrsel(uint32 enabled)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_S2CSEL_REG,RTL8368S_SVLAN_S2CSEL_OFF,enabled);

	return retVal;
}


/*
@func int32 | rtl8368s_getAsicSvlanS2Cmbrsel | Configure enable S-VLAN  function.
@parm uint32 | enabled | S-VLAN S2C member selection function 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	 There are service provider VLAN functions defined in IEEE 802.1ad standard. System supports user defined type/length for parsing S-tag frame at uplink 
	 ports. Without S-VLAN function enable setting, S-tag frame from uplink ports will be treated as normal frame payload to forward. Only S-tag frame contained
	 supported S-VID will be accepted from uplink ports.
*/
int32 rtl8368s_getAsicSvlanS2Cmbrsel(uint32* enabled)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_S2CSEL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_S2CSEL_MSK;

	if(regData)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicSvlanPrioDecision | Configure priority decision of S-tag frame output to uplink ports.
@parm uint32 | vsPrio | Priority of S-tag decision, 0: usge C-Tag's priority as S-priority 1: use mapped priority indexed from Pn_VSIDX.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	If setting is 0 by using C-tag's priority, then two situations need to be take care. Original receiving frame is C-tag frame and S-priority will be set as
	same as C-priority in C-tag. Otherwise, S-priority of out frame to uplink ports should be user defined S-priority indirectly mapping at Pn_VSIDX.
	
*/
int32 rtl8368s_setAsicSvlanPrioDecision(uint32 vsPrio)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_PRIO_OFF,vsPrio);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanPrioDecision | Get priority decision of S-tag frame output to uplink ports.
@parm uint32* | vsPrio | Priority of S-tag decision, 0: usge C-Tag's priority as S-priority 1: use mapped priority indexed from Pn_VSIDX.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get output priority decision method while sending s-tag frame to uplink port
	
*/
int32 rtl8368s_getAsicSvlanPrioDecision(uint32* vsPrio)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_PRIO_MSK;

	if(regData)
		*vsPrio = 1;
	else
		*vsPrio = 0;

	return SUCCESS;
	
}

/*
@func int32 | rtl8368s_setAsicSvlanTag | Configure if output S-tag frame to uplink ports. 
@parm uint32 | enabled | Output frame to uplink ports 0: S-untag 1:S-tag 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	If register VS_STAG is set as 0, frames transmited through uplink port should be S-untagged.Otherwise, it should be S-tagged frames. Both S-VID and
	priority are defined from indirectly indexing at Pn_VSIDX at each uplink port. System supports 4 accepted S-VIDs only. In spec. 802.1ad, output frame 
	should be S-tag frame only (S-tag frames are transferred between customer bridge and priovider bridge).
*/
int32 rtl8368s_setAsicSvlanTag(uint32 enabled)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_STAG_OFF,enabled);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanTag | Get system output S-tag frame to uplink ports. 
@parm uint32* | enabled | Output frame to uplink ports 0: S-untag 1:S-tag 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get setting that system outputs S-tag frame to uplink ports or not.
*/
int32 rtl8368s_getAsicSvlanTag(uint32* enabled)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_STAG_MSK;

	if(regData)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;	
}

/*
@func int32 | rtl8368s_setAsicSvlanUplinkPortMask | Configure uplink ports mask.
@parm uint32 | portMask | Uplink port mask setting.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	Uplink port mask is setting which ports are connected to provider switch. If ports are belong uplink ports and all frames receiving from these port must 
	contain accept SVID in S-tag field.
*/
int32 rtl8368s_setAsicSvlanUplinkPortMask(uint32 portMask)
{
	uint32 retVal;
	uint32 regData;

	if(portMask > RTL8368S_PORTMASK)
		return ERRNO_INPUT;

	regData = portMask << RTL8368S_SVLAN_PORT_OFF;

	retVal = rtl8368s_setAsicRegBits(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_PORT_MSK,regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanUplinkPortMask | Get uplink ports mask configuration.
@parm uint32* | portMask | Uplink port mask setting.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get setting of  uplink port mask belong to service provider VLAN ports.
	
*/
int32 rtl8368s_getAsicSvlanUplinkPortMask(uint32* portMask)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;


	regData = (regData& RTL8368S_SVLAN_PORT_MSK ) >> RTL8368S_SVLAN_PORT_OFF;

	*portMask = regData;

	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicSvlanProtocolType | Configure accepted S-VLAN ether type. The default ether type of S-VLAN is 0x88a8.
@parm uint32 | protocolType | Ether type of S-tag frame parsing in uplink ports.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	Ether type of S-tag in 802.1ad is 0x88a8 and there are existed ether type 0x9100 and 0x9200 for Q-in-Q SLAN design. User can set mathced ether
	type as service provider supported protocol. 

*/
int32 rtl8368s_setAsicSvlanProtocolType(uint32 protocolType)
{
	uint32 retVal;


	retVal = rtl8368s_setAsicReg(RTL8368S_SVLAN_PROTOCOL_TYPE,protocolType);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanProtocolType | Get accepted S-VLAN ether type setting.
@parm uint32 | protocolType |  Ether type of S-tag frame parsing in uplink ports.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get ccepted ether type of service provider VLAN tag. The default ether type of service provider VLAN in 802.1ad is 0x88a8.

*/
int32 rtl8368s_getAsicSvlanProtocolType(uint32* protocolType)
{
	uint32 retVal;
	uint32 regData;


	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_PROTOCOL_TYPE, &regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*protocolType = regData;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicSvlanConfiguration| Configure system 8 S-tag content
@parm uint32 | index | index of 8 s-tag configuration
@parm unit32 | svid | SVLAN VID
@parm enum PRIORITYVALUE | spri | Priority S-tag.
@parm unit32 | rpvid | Replace PVID of un-Ctagging frame.
@parm unit32 | cvidx | Cvidx for replace PVID
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can set system 4 accepted s-tag frame format. Only 4 SVID s-tag frame will be accpeted
	to receiving from uplink ports. Other SVID s-tag frame or s-utagged frame will be droped in uplink ports.

*/
int32 rtl8368s_setAsicSvlanConfiguration(uint32 index,uint32 svid,enum PRIORITYVALUE spri,uint32 rpvid,uint32 cvidx)
{
	uint32 retVal;
	uint16 regAddr;
	uint32 regData;
	uint32 regBits;


	if(index > RTL8368S_SVIDXMAX)
		return ERRNO_SVIDX;

	if(svid > 0xFFF)
		return ERRNO_VID;

	if(spri >= PRI_MAX)
		return ERRNO_PRIORITY;

	if(cvidx > RTL8368S_VIDXMAX)
		return ERRNO_CVIDX;

	regData = ((svid<<RTL8368S_SVLAN_SVIDX_VID_OFF)&RTL8368S_SVLAN_SVIDX_VID_MSK) |
			  ((spri<<RTL8368S_SVLAN_SVIDX_SPRI_OFF)&RTL8368S_SVLAN_SVIDX_SPRI_MSK) |
			  ((rpvid<<RTL8368S_SVLAN_SVIDX_CRPL_OFF)&RTL8368S_SVLAN_SVIDX_CRPL_MSK);

	retVal = rtl8368s_setAsicReg(RTL8368S_SVLAN_SVIDX_BASE+index,regData);
	if(SUCCESS != retVal)
		return retVal;

	if(index < 4)
	{
		regAddr = RTL8368S_SVLAN_CVIDX_BASE;
		regBits = RTL8368S_SVLAN_CVIDX_MSK<<(index*RTL8368S_SVLAN_CVIDX_BITS);
		regData = (cvidx<<(index*RTL8368S_SVLAN_CVIDX_BITS)) & regBits;
	}	
	else
	{
		regAddr = RTL8368S_SVLAN_CVIDX_BASE + 1;
		regBits = RTL8368S_SVLAN_CVIDX_MSK<<((index-4)*RTL8368S_SVLAN_CVIDX_BITS);
		regData = (cvidx<<((index-4)*RTL8368S_SVLAN_CVIDX_BITS)) & regBits;
	}	

	
	retVal = rtl8368s_setAsicRegBits(regAddr,regBits,regData);		
	
	return retVal;
	
}	

/*
@func int32 | rtl8368s_getAsicSvlanConfiguration| Configure system 8 S-tag content
@parm uint32* | index | index of 8 s-tag configuration
@parm unit32* | svid | SVLAN VID
@parm enum PRIORITYVALUE* | spri | Priority S-tag.
@parm unit32* | rpvid | Replace PVID of un-Ctagging frame.
@parm unit32* | cvidx | Cvidx for replace PVID
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get system 4 accepted s-tag frame format. Only 4 SVID s-tag frame will be accpeted
	to receiving from uplink ports. Other SVID s-tag frame or s-utagged frame will be droped in uplink ports.

*/
int32 rtl8368s_getAsicSvlanConfiguration(uint32 index,uint32* svid,enum PRIORITYVALUE* spri,uint32* rpvid,uint32* cvidx)
{

	uint32 retVal;
	uint16 regAddr;
	uint32 regData;
	uint32 regBits;
	uint32 bitsOff;

	if(index > RTL8368S_SVIDXMAX)
		return ERRNO_SVIDX;


	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_SVIDX_BASE+index,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	*svid = (regData & RTL8368S_SVLAN_SVIDX_VID_MSK) >> RTL8368S_SVLAN_SVIDX_VID_OFF;
	*spri = (regData & RTL8368S_SVLAN_SVIDX_SPRI_MSK) >> RTL8368S_SVLAN_SVIDX_SPRI_OFF;

	if(regData & RTL8368S_SVLAN_SVIDX_CRPL_MSK) 
		* rpvid = TRUE;
	else
		* rpvid = FALSE;

	if(index < 4)
	{
		regAddr = RTL8368S_SVLAN_CVIDX_BASE;
		regBits = RTL8368S_SVLAN_CVIDX_MSK<<(index*RTL8368S_SVLAN_CVIDX_BITS);
		bitsOff = index*RTL8368S_SVLAN_CVIDX_BITS;
	}	
	else
	{
		regAddr = RTL8368S_SVLAN_CVIDX_BASE + 1;
		regBits = RTL8368S_SVLAN_CVIDX_MSK<<((index-4)*RTL8368S_SVLAN_CVIDX_BITS);
		bitsOff = (index-4)*RTL8368S_SVLAN_CVIDX_BITS;
	}	

	retVal = rtl8368s_getAsicReg(regAddr,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*cvidx =  (regData & regBits) >> bitsOff;

	return SUCCESS;
}
/*
@func int32 | rtl8368s_setAsicSvlanPortVsidx| Configure per port stag indirect index to 4 stag configuration
@parm enum PORTID | port | Physical port number.
@parm uint32 | vsidx | index of 4 s-tag configuration
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid Port Number.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can set port n S-tag format index while receiving frame from port n 
	is transmit through uplink port with s-tag field

*/
int32 rtl8368s_setAsicSvlanPortVsidx(enum PORTID port, uint32 vsidx)
{
	uint32 regData;
	uint32 regBits;
	uint32 regAddr;
	int32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	if(vsidx > RTL8368S_SVIDXMAX )
		return ERRNO_SVIDX;



	if(port < PORT5)
	{
		regAddr = RTL8368S_SVLAN_PORT_CTRL_BASE;

		regBits = RTL8368S_SVLAN_VSIDX_MSK<<(RTL8368S_SVLAN_VSIDX_BITS*port);

		regData = (vsidx << (RTL8368S_SVLAN_VSIDX_BITS*port)) & regBits;
	}
	else
	{
		regAddr = RTL8368S_SVLAN_PORT_CTRL_BASE + 1;

		regBits = RTL8368S_SVLAN_VSIDX_MSK<<(RTL8368S_SVLAN_VSIDX_BITS*(port-PORT5));

		regData =(vsidx << (RTL8368S_SVLAN_VSIDX_BITS*(port-PORT5))) & regBits;
	}

	retVal = rtl8368s_setAsicRegBits(regAddr,regBits,regData);		

	return retVal;	
}

/*
@func int32 | rtl8368s_getAsicSvlanPortVsidx| get per port stag indirect index to 4 stag configuration
@parm enum PORTID | port | Physical port number.
@parm uint32* | vsidx | index of 4 s-tag configuration
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid Port Number.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get port n S-tag format index while receiving frame from port n 
	is transmit through uplink port with s-tag field

*/
int32 rtl8368s_getAsicSvlanPortVsidx(enum PORTID port, uint32* vsidx)
{
	int32 retVal;
	uint32 regData;
	uint32 regBits;
	uint32 bitsOff;
	uint32 regAddr;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;


	if(port < PORT5)
	{
		regAddr = RTL8368S_SVLAN_PORT_CTRL_BASE;

		regBits = RTL8368S_SVLAN_VSIDX_MSK<<(RTL8368S_SVLAN_VSIDX_BITS*port);

		bitsOff = RTL8368S_SVLAN_VSIDX_BITS*port;
	}	
	else
	{
		regAddr = RTL8368S_SVLAN_PORT_CTRL_BASE + 1;

		regBits = RTL8368S_SVLAN_VSIDX_MSK<<(RTL8368S_SVLAN_VSIDX_BITS*(port-PORT5));

		bitsOff = RTL8368S_SVLAN_VSIDX_BITS*(port-PORT5);
	}	

	retVal = rtl8368s_getAsicReg(regAddr,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*vsidx =  (regData & regBits) >> bitsOff;

	return SUCCESS;	
}


/*
@func int32 | rtl8368s_setAsicSvlanIngressUntag | Configure enable trap received un-Stag frame to CPU from unplink port/
@parm uint32 | enabled | Trap received un-Stag frame to CPU 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	 The API can set trapping function in uplink port while receiving un-Stagging frame.
*/
int32 rtl8368s_setAsicSvlanIngressUntag(uint32 enabled)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_ING_UNTAG_OFF,enabled);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanIngressUntag | Configure enable trap received un-Stag frame to CPU from unplink port.
@parm uint32* | enabled | Trap received un-Stag frame to CPU 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get configuration of trapping function in uplink port while receiving un-Stagging frame.
	
*/
int32 rtl8368s_getAsicSvlanIngressUntag(uint32* enabled)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_ING_UNTAG_MSK;

	if(regData)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;
	
}


/*
@func int32 | rtl8368s_setAsicSvlanIngressUnmatch | Configure enable trap received unmatched Stag frame to CPU from unplink port.
@parm uint32 | enabled | Trap received unmatched Stag frame to CPU 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	 The API can set trapping function in uplink port while receiving unmatched Stag frame.
*/
int32 rtl8368s_setAsicSvlanIngressUnmatch(uint32 enabled)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_ING_UNMAT_OFF,enabled);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanIngressUnmatch | Configure enable trap received unmatched Stag frame to CPU from unplink port.
@parm uint32* | enabled | Trap received unmatched Stag frame to CPU 1: enabled, 0: disabled. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get configuration of trapping function in uplink port while receiving unmatched Stag frame.
	
*/
int32 rtl8368s_getAsicSvlanIngressUnmatch(uint32* enabled)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;

	regData = regData & RTL8368S_SVLAN_ING_UNMAT_MSK;

	if(regData)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;
	
}


/*
@func int32 | rtl8368s_setAsicSvlanCpuPriority | Assign CPU priority for SVLAN trapping frame.
@parm enum PRIORITYVALUE | priority | CPU priority for trapping frame.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can assign CPU priority for unmacthed and unstag frame were trapped to CPU and this priority will be remapping after qos priority decison.
*/
int32 rtl8368s_setAsicSvlanCpuPriority(enum PRIORITYVALUE priority)
{
	uint32 retVal;
	uint32 regData;

	if(priority > PRI7)
		return ERRNO_PRIORITY; 

	regData = priority << RTL8368S_SVLAN_CPUPRI_OFF;

	retVal = rtl8368s_setAsicRegBits(RTL8368S_SVLAN_CTRL_REG,RTL8368S_SVLAN_CPUPRI_MSK,regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSvlanCpuPriority | Get CPU priority for SVLAN trapping frame.
@parm enum PRIORITYVALUE* | priority | CPU priority for trapping frame.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can retrieve CPU priority for unmacthed and unstag frame were trapped to CPU
	
*/
int32 rtl8368s_getAsicSvlanCpuPriority(enum PRIORITYVALUE* priority)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_SVLAN_CTRL_REG,&regData);

	if(retVal !=  SUCCESS)
		return retVal;


	regData = (regData& RTL8368S_SVLAN_CPUPRI_MSK ) >> RTL8368S_SVLAN_CPUPRI_OFF;

	*priority = regData;

	return SUCCESS;
}




/*
@func int32 | rtl8368s_setAsicCpuPortMask| Configure cpu ports
@parm uint32 | portMask | Uplink port mask setting.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can specify cpu port. All the frames send to cpu port will be
	inserted proprietary cpu tag 0x8899 if NOT disable insert CPU tag function.

*/
int32 rtl8368s_setAsicCpuPortMask(uint32 portMask)
{
	int32 retVal;


	retVal = rtl8368s_setAsicRegBits(RTL8368S_CPU_CTRL_REG,RTL8368S_CPU_PORTS_MSK,portMask<<RTL8368S_CPU_PORTS_OFF);		

	return retVal;	
}

/*
@func int32 | rtl8368s_getAsicCpuPortMask| Get cpu ports
@parm uint32* | portMask | Uplink port mask setting.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can check if the specified port is a CPU port.

*/
int32 rtl8368s_getAsicCpuPortMask(uint32* portMask)
{
	uint32 regData;
	int32 retVal;

	retVal = rtl8368s_getAsicReg(RTL8368S_CPU_CTRL_REG, &regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*portMask = (regData & RTL8368S_CPU_PORTS_MSK) >> RTL8368S_CPU_PORTS_OFF;

	return SUCCESS;
}
/*
@func int32 | rtl8368s_setAsicCpuDisableInsTag | Set CPU port DISABLE insert tag function
@parm uint32 | enable | 0: disable, 1: enable. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	Enable the function to NOT insert proprietary CPU tag (Length/Type 0x8899) ahead vlan tag
	to the frame that transmitted to CPU port or disable the function to insert proprietary CPU tag.
	
*/
int32 rtl8368s_setAsicCpuDisableInsTag(uint32 enable)
{
	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_CPU_CTRL_REG,RTL8368S_CPU_INSTAG_OFF,enable);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicCpuDisableInsTag | Get setting of CPU port DISABLE insert tag function
@parm uint32* | enabled | 0: disable, 1: enable. 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid input parameter.
@comm
	The API can get the setting of CPU port DISABLE insert tag function.
	
*/
int32 rtl8368s_getAsicCpuDisableInsTag(uint32* enabled)
{
	uint32 regData;
	int32 retVal;

	retVal = rtl8368s_getAsicReg(RTL8368S_CPU_CTRL_REG, &regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData & RTL8368S_CPU_INSTAG_MSK)
		*enabled = 1;
	else
		*enabled = 0;
	
	return SUCCESS;
}


/*
@func int32 | rtl8368s_setAsicCpuPriorityRemapping | Set CPU priority remapping parameter.
@parm enum PRIORITYVALUE | priority | Priority value.
@parm enum PRIORITYVALUE | newpriority | New priority value.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PRIORITY | Invalid priority. 
@comm
	The API can set CPU reampping source priority and new priority.
 */
int32 rtl8368s_setAsicCpuPriorityRemapping( enum PRIORITYVALUE priority, enum PRIORITYVALUE newpriority )
{
	uint32 retVal;
	uint16 regAddr;
	uint32 regData;
	uint32 regBits;
	
	/* Invalid input parameter */
	if ((priority < PRI0) || (priority > PRI7) || (newpriority < PRI0) || (newpriority > PRI7))
		return ERRNO_PRIORITY; 


	if(priority <= PRI4)
	{
		regAddr = RTL8368S_CPU_PRIORITY_BASE;
		regBits = RTL8368S_CPU_PRIORITY_MSK<<(priority*RTL8368S_CPU_PRIORITY_BITS);
		regData = (newpriority<<(priority*RTL8368S_CPU_PRIORITY_BITS)) & regBits;
	}	
	else
	{
		regAddr = RTL8368S_CPU_PRIORITY_BASE + 1;
		regBits = RTL8368S_CPU_PRIORITY_MSK<<((priority-PRI5)*RTL8368S_CPU_PRIORITY_BITS);
		regData = (newpriority<<((priority-PRI5)*RTL8368S_CPU_PRIORITY_BITS)) & regBits;
	}	

	
	retVal = rtl8368s_setAsicRegBits(regAddr,regBits,regData);		
	
	return retVal;
}

/*
@func int32 | rtl8368s_getAsicCpuPriorityRemapping | Get CPU priority remapping parameter.
@parm enum PRIORITYVALUE | priority | Priority value.
@parm enum PRIORITYVALUE* | newpriority | New priority value.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PRIORITY | Invalid priority. 
@comm
	The API can get CPU reampping source priority and new priority.
 */
int32 rtl8368s_getAsicCpuPriorityRemapping( enum PRIORITYVALUE priority, enum PRIORITYVALUE *pNewpriority )
{
	uint32 retVal;
	uint16 regAddr;
	uint32 regData;
	uint32 regBits;
	uint32 bitsOff;

	/* Invalid input parameter */
	if ((priority < PRI0) || (priority > PRI7))
		return ERRNO_PRIORITY; 

	if(priority <= PRI4)
	{
		regAddr = RTL8368S_CPU_PRIORITY_BASE;
		regBits = RTL8368S_CPU_PRIORITY_MSK<<(priority*RTL8368S_CPU_PRIORITY_BITS);
		bitsOff = priority*RTL8368S_CPU_PRIORITY_BITS;
	}	
	else
	{
		regAddr = RTL8368S_CPU_PRIORITY_BASE + 1;
		regBits = RTL8368S_CPU_PRIORITY_MSK<<((priority-PRI5)*RTL8368S_CPU_PRIORITY_BITS);
		bitsOff = (priority-PRI5)*RTL8368S_CPU_PRIORITY_BITS;
	}	

	retVal = rtl8368s_getAsicReg(regAddr,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*pNewpriority =  (regData & regBits) >> bitsOff;
	return SUCCESS;

}



/*
@func int32 | rtl8368s_setAsicLinkAggregationMode | Set link aggregation mode.
@parm uint32 | mode | Link aggragation mode 1:dump 0:not dump.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	If software set link aggragation mode to dump, then ASIC will automatic set hash value mapping to link aggragation ports.  For example, aggregation port mask
	is set as 0x001011, which means that port 0,1,3 are aggregating ports, and hash value from 0 to 7 will be mapped to port (0,1,3,0,1,3,0,1). Before changing link
	aggragation mode, system should make sure that output queues in aggregated ports are empty to avoid out of order packets issuse.
*/
int32 rtl8368s_setAsicLinkAggregationMode(uint32 mode)
{

	uint32 retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_3AD_LINKAG_CTRL_REG,RTL8368S_3AD_LINKAG_MODE_OFF,mode);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationMode | Get link aggregation mode configuration.
@parm uint32* | mode | Link aggragation mode 1:dump 0:not dump.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can get Link Aggragation mode. 
*/
int32 rtl8368s_getAsicLinkAggregationMode(uint32 *mode)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_3AD_LINKAG_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData & RTL8368S_3AD_LINKAG_MODE_MSK)
		*mode = 1;
	else
		*mode = 0;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicLinkAggregationHashAlgorithm | Configure link aggragation hashing algorithm.
@parm uint32 | algorithm | Link aggregation algorithm 0:hash(DA+SA) 1:hash(DA) 2:hash(SA).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_LA_INVALIDHASHSEL | Invalid link aggregation hash algorithm selection.
@comm
	Either link aggragation dump mode or not dump mode, the hash algorithm will follow user setting. The hash algorithm value 00 means using hash(DA+SA)
	algorithm, value 01 means using hash(DA) algorithm and value 10 means using hash(SA) algorithm. Before changing link	aggragation algorithm, system should
	make sure that output queues in aggregated ports are empty to avoid out of order packets issuse.
*/
int32 rtl8368s_setAsicLinkAggregationHashAlgorithm(uint32 algorithm)
{
	uint32 retVal;
	uint32 regData;

	if(algorithm > RTL8368S_LAHASHSELMAX)
		return ERRNO_LA_INVALIDHASHSEL;

	regData = algorithm << RTL8368S_3AD_LINKAG_HASHSEL_OFF;

	retVal = rtl8368s_setAsicRegBits(RTL8368S_3AD_LINKAG_CTRL_REG,RTL8368S_3AD_LINKAG_HASHSEL_MSK,regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationHashAlgorithm | Get link aggregation hashing algorithm.
@parm uint32 | algorithm | Link aggregation algorithm 0:hash(DA+SA) 1:hash(DA) 2:hash(SA).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can get link aggregation hash algorithm.

*/
int32 rtl8368s_getAsicLinkAggregationHashAlgorithm(uint32 *algorithm)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_3AD_LINKAG_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*algorithm = (regData&RTL8368S_3AD_LINKAG_HASHSEL_MSK)>>RTL8368S_3AD_LINKAG_HASHSEL_OFF;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicLinkAggregationPortMask | set link aggragatio port mask
@parm uint32 | portmask | support maximum 4 physicial ports
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_LA_INVALIDPORTS | invalid link aggragation port numbers.
@comm
	The API can set Link Aggragation port mask. Asic supports only 4 aggregation ports.

*/
int32 rtl8368s_setAsicLinkAggregationPortMask(uint32 portmask)
{
	uint32 retVal;
	uint32 regData;
	uint32 maskportNo;
	uint32 i;

	maskportNo = 0;
	/*checking aggregation port number*/
	/*ASIC support maximum 4 ports aggregation function*/
	for(i=0;i<PORT_MAX;i++)
	{
		if((0x01<<i) & portmask)	
			maskportNo ++;
	}

	if(maskportNo > RTL8368S_LAPORTSMAX)
		return ERRNO_LA_INVALIDPORTS;

	regData = portmask << RTL8368S_3AD_LINKAG_PORTMASK_OFF;

	regData = regData & RTL8368S_3AD_LINKAG_PORTMASK_MSK;

	retVal = rtl8368s_setAsicRegBits(RTL8368S_3AD_LINKAG_CTRL_REG,RTL8368S_3AD_LINKAG_PORTMASK_MSK,regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationPortMask | set link aggragatio port mask
@parm uint32* | portmask | support maximum 4 physicial ports
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set Link Aggragation port mask. Asic supports only 4 aggregation ports.

*/
int32 rtl8368s_getAsicLinkAggregationPortMask(uint32* portmask)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_3AD_LINKAG_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*portmask = (regData&RTL8368S_3AD_LINKAG_PORTMASK_MSK)>>RTL8368S_3AD_LINKAG_PORTMASK_OFF;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicLinkAggregationHashTable | set link aggragatio hash value mapping port number
@parm uint32 | hashval | hashing value 0-7
@parm enum PORTID | port | Physical port number.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_LA_INVALIDHASHVAL | Invalid hash value. 0-7 only
@rvalue ERRNO_INPUT | Invalid Port Number.
@comm
	The API can set Link Aggragation hash value mapped to port ID. Hash value range is 0-7 and port range is 0-5.
	If mapped port ID was not set as aggragated port and ASIC will still follow the mapping result to send frame to
	port as hashing result mapping.

*/
int32 rtl8368s_setAsicLinkAggregationHashTable(uint32 hashval,enum PORTID port)
{
	uint32 retVal;
	uint32 regData;
	uint32 regAddr;
	uint32 regBits;


	/* bits mapping to hash value */
	const uint16 bits[8]= { 0x0007,0x0038,0x01C0,0x0E00,0x7000,0x0007,0x0038,0x01C0 };
	/* bits offset to hash value*/
	const uint16 bitOff[8] = { 0,3,6,9,12,0,3,6 };
	/* address offset to hash value */
	const uint16 addrOff[8]= { 0,0,0,0,0,1,1,1 };


	if(hashval > RTL8368S_LAHASHVALMAX)
		return ERRNO_LA_INVALIDHASHVAL;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	regAddr = RTL8368S_3AD_LINKAG_MAPPING_BASE + addrOff[hashval];

	regBits = bits[hashval];

	regData =  (port << bitOff[hashval]) & regBits;

	retVal = rtl8368s_setAsicRegBits(regAddr,regBits,regData);		
	
	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationHashTable | get link aggragatio hash value mapping port number
@parm uint32 | hashval | hashing value 0-7
@parm enum PORTID* | port | Physical port number.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_LA_INVALIDHASHVAL | Invalid hash value. 0-7 only
@comm
	The API can get Link Aggragation hash value mapped to port ID. Hash value range is 0-7 and port range is 0-5.
	If mapped port ID was not set as aggragated port and ASIC will still follow the mapping result to send frame to
	port as hashing result mapping.

*/
int32 rtl8368s_getAsicLinkAggregationHashTable(uint32 hashval,enum PORTID* port)
{
	uint32 retVal;
	uint32 regData;
	uint32 regAddr;


	/* bits mapping to hash value */
	const uint16 bits[8]= { 0x0007,0x0038,0x01C0,0x0E00,0x7000,0x0007,0x0038,0x01C0 };
	/* bits offset to hash value*/
	const uint16 bitOff[8] = { 0,3,6,9,12,0,3,6 };
	/* address offset to hash value */
	const uint16 addrOff[8]= { 0,0,0,0,0,1,1,1 };


	if(hashval > RTL8368S_LAHASHVALMAX)
		return ERRNO_LA_INVALIDHASHVAL;



	regAddr = RTL8368S_3AD_LINKAG_MAPPING_BASE + addrOff[hashval];

	retVal = rtl8368s_getAsicReg(regAddr,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*port =  (regData & bits[hashval]) >> bitOff[hashval];


	return retVal;
}

/*
@func int32 | rtl8368s_setAsicLinkAggregationFlowControl | set link aggragatio flow control
@parm uint32 | fcport | port mask for flow control of link aggragation
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set port mask for flow control in Link Aggragation

*/
int32 rtl8368s_setAsicLinkAggregationFlowControl(uint32 fcport)
{
	uint32 retVal;
	uint32 regData;

	regData = fcport & RTL8368S_3AD_LINKAG_FC_MSK;

	retVal = rtl8368s_setAsicReg(RTL8368S_3AD_LINKAG_FC_CTRL_REG,regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationFlowControl | get link aggragatio flow control
@parm uint32* | fcport | port mask for flow control of link aggragation
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set port mask for flow control in Link Aggragation

*/
int32 rtl8368s_getAsicLinkAggregationFlowControl(uint32* fcport)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_3AD_LINKAG_FC_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*fcport = regData&RTL8368S_3AD_LINKAG_FC_MSK;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_getAsicLinkAggregationQueueEmpty | get link aggragation queue empty status
@parm uint32* | qeport | port mask of queue status
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can get port mask for queueing packet empty status for each port

*/
int32 rtl8368s_getAsicLinkAggregationQueueEmpty(uint32* qeport)
{
	uint32 retVal;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_3AD_LINKAG_QEMPTY_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*qeport = regData&RTL8368S_3AD_LINKAG_QEMPTY_MSK;
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicStormFiltering | Set per port  storm filtering function.
@parm enum PORTID | port | Physical Port number.
@parm uint32 | bcstorm | broadcasting storm filtering 1:enable, 0:disabled.
@parm uint32 | mcstorm | multicasting storm filtering 1:enable, 0:disabled.
@parm uint32 | undastorm | unknown destination storm filtering 1:enable, 0:disabled.
@parm uint32 | unmcstorm | unknown multicast storm filtering 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error. 
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@comm
	The API can be used to enable or disable 'Storm Filtering Control'. 
*/
int32 rtl8368s_setAsicStormFiltering(enum PORTID port, uint32 bcstorm, uint32 mcstorm, uint32 undastorm,uint32 unmcstorm)
{
	uint32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	/*broadcasting storm filtering control*/
	retVal = rtl8368s_setAsicRegBit(RTL8368S_STORM_BC_CTRL_REG,port,bcstorm);
	if(retVal !=  SUCCESS)
		return retVal;

	/*multicasting storm filtering control*/
	retVal = rtl8368s_setAsicRegBit(RTL8368S_STORM_MC_CTRL_REG,port,mcstorm);
	if(retVal !=  SUCCESS)
		return retVal;

	/*unknown destination storm filtering control*/
	retVal = rtl8368s_setAsicRegBit(RTL8368S_STORM_UNDA_CTRL_REG,port,undastorm);
	if(retVal !=  SUCCESS)
		return retVal;

	/*unknown multicast storm filtering control*/
	retVal = rtl8368s_setAsicRegBit(RTL8368S_STORM_UNMC_CTRL_REG,port,unmcstorm);
	if(retVal !=  SUCCESS)
		return retVal;

	return SUCCESS;
}

/*
@func int32 | rtl8368s_getAsicStormFiltering | get per port  storm filtering function.
@parm enum PORTID | port | Physical Port number.
@parm uint32* | bcstorm | broadcasting storm filtering 1:enable, 0:disabled.
@parm uint32* | mcstorm | multicasting storm filtering 1:enable, 0:disabled.
@parm uint32* | undastorm | unknown destination storm filtering 1:enable, 0:disabled.
@parm uint32* | unmcstorm | unknown multicast storm filtering 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error. 
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@comm
	The API can be used to enable or disable 'Storm Filtering Control'. 
*/
int32 rtl8368s_getAsicStormFiltering(enum PORTID port, uint32* bcstorm, uint32* mcstorm, uint32* undastorm,uint32* unmcstorm)
{
	uint32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;


	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_BC_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData&(1<<port))
		*bcstorm = 1;
	else
		*bcstorm = 0;

	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_MC_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData&(1<<port))
		*mcstorm = 1;
	else
		*mcstorm = 0;

	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_UNDA_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData&(1<<port))
		*undastorm = 1;
	else
		*undastorm = 0;

	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_UNMC_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData&(1<<port))
		*unmcstorm = 1;
	else
		*unmcstorm = 0;

	return SUCCESS;
}
/*
@func int32 | rtl8368s_setAsicStormFiltering | Set system storm filtering meter.
@parm enum PORTID | port | Physical Port number.
@parm uint32 | rate | storm filtering rate 0~0x3FFF. (uint 64kpbs)
@parm uint32 | ifg | storm filtering rate's calculation including IPG 1:include 0:exclude 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error. 
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@comm
	The API can be used to set per system storm filtering meter
*/
int32 rtl8368s_setAsicStormFilteringMeter(enum PORTID port,uint32 rate, uint32 ifg )
{
	uint32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;


	retVal = rtl8368s_setAsicReg(RTL8368S_STORM_RATE_CTRL_BASE+port,rate&RTL8368S_STORM_RATE_MSK);
	if(retVal !=  SUCCESS)
		return retVal;

	retVal = rtl8368s_setAsicRegBit(RTL8368S_STORM_MISC_CTRL_REG,RTL8368S_STORM_MISC_IFG_OFF,ifg);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicStormFiltering | get system storm filtering meter.
@parm enum PORTID | port | Physical Port number.
@parm uint32* | rate | storm filtering rate 0~0x3FFF. (uint 64kpbs)
@parm uint32* | ifg | storm filtering rate's calculation including IPG 1:include 0:exclude 
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error. 
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@comm
	The API can be used to get per system storm filtering meter
*/
int32 rtl8368s_getAsicStormFilteringMeter(enum PORTID port,uint32* rate, uint32* ifg )
{
	uint32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_RATE_CTRL_BASE+port,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	* rate = regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_MISC_CTRL_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData & RTL8368S_STORM_MISC_IFG_MSK)
	{
		*ifg = 1;
	}
	else
	{
		*ifg = 0;
	}
		
	return SUCCESS;
}

/*
@func int32 | rtl8368s_getAsicStormFilteringExceedStatus | get system storm filtering meter over status.
@parm uint32* | portmask | Port mask for storm control rate have been exceeded status.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error. 
@comm
	The API can be get storm control filtering function reach the limitation status. If the port rate limitation had been exceeded, then port
	mapped status will be set.
*/
int32 rtl8368s_getAsicStormFilteringExceedStatus(uint32* portmask)
{
	uint32 retVal;
	uint32 regData;


	retVal = rtl8368s_getAsicReg(RTL8368S_STORM_LEAKY_STATUS_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*portmask = (regData & RTL8368S_STORM_EXCEED_MSK) >> RTL8368S_STORM_EXCEED_OFF;

	return SUCCESS;
}

/*=======================================================================
 *PHY APIs 
 *========================================================================*/
/*
@func int32 | rtl8368s_setAsicPHYReg | Set PHY registers .
@parm uint32 | phyNo | PHY number (0~3).
@parm uint32 | page | PHY page (0~7).
@parm uint32 | addr | PHY address (0~31).
@parm uint32 | data | Writing data.
@rvalue SUCCESS | 
@rvalue FAILED | invalid parameter
@comm
 */
int32 rtl8368s_setAsicPHYRegs( uint32 phyNo, uint32 page, uint32 addr, uint32 data )
{
	int32   retVal,phySmiAddr;
    uint32  dummy;

	if(phyNo > RTL8368S_PHY_NO_MAX)
		return ERROR_PHY_INVALIDPHYNO;

	if(page > RTL8368S_PHY_PAGE_MAX)
		return ERROR_PHY_INVALIDPHYPAGE;

	if(addr > RTL8368S_PHY_REG_MAX)
		return ERROR_PHY_INVALIDREG;

    /* Dummy read before write PHY register */
    retVal = rtl8368s_getAsicPHYRegs(phyNo, page, addr, &dummy);
    if (retVal !=  SUCCESS) 
		return retVal;

	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_WRITE);
	if (retVal !=  SUCCESS) 
		return retVal;

	phySmiAddr = 0x8000 | (1<<(phyNo +RTL8368S_PHY_NO_OFFSET)) | 
			((page <<RTL8368S_PHY_PAGE_OFFSET)&RTL8368S_PHY_PAGE_MASK) | (addr &RTL8368S_PHY_REG_MASK);
	
	retVal = rtl8368s_setAsicReg(phySmiAddr, data);
	if (retVal !=  SUCCESS) 
		return retVal;


	return SUCCESS;	
}

/*
@func int32 | rtl8368s_getAsicPHYReg | get PHY registers .
@parm uint32 | phyNo | PHY number (0~3).
@parm uint32 | page | PHY page (0~7).
@parm uint32 | addr | PHY address (0~31).
@parm uint32* | data | Read data.
@rvalue SUCCESS | 
@rvalue FAILED | invalid parameter
@comm
 */
int32 rtl8368s_getAsicPHYRegs( uint32 phyNo, uint32 page, uint32 addr, uint32 *data )
{
	int32 retVal,phySmiAddr;

	if(phyNo > RTL8368S_PHY_NO_MAX)
		return ERROR_PHY_INVALIDPHYNO;

	if(page > RTL8368S_PHY_PAGE_MAX)
		return ERROR_PHY_INVALIDPHYPAGE;

	if(addr > RTL8368S_PHY_REG_MAX)
		return ERROR_PHY_INVALIDREG;

	
	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_READ);
	if (retVal !=  SUCCESS) 
		return retVal;

	phySmiAddr = 0x8000 | (1<<(phyNo +RTL8368S_PHY_NO_OFFSET)) | 
			((page <<RTL8368S_PHY_PAGE_OFFSET)&RTL8368S_PHY_PAGE_MASK) | (addr &RTL8368S_PHY_REG_MASK);
	
	retVal = rtl8368s_setAsicReg(phySmiAddr, 0);
	if (retVal !=  SUCCESS) 
		return retVal;

	//For Verify PHY Read issue, mark this place
	retVal = rtl8368s_setAsicReg(phySmiAddr, 0);
	if (retVal !=  SUCCESS) 
		return retVal;

	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_DATA_REG,data);
	if(retVal !=  SUCCESS)
		return retVal;

	return SUCCESS;	
}
/*
@func int32 | rtl8368s_setAsicExtPHYReg | Set external PHY registers .
@parm uint32 | phy | PHY address (0~31).
@parm uint32 | reg | PHY register (0~31).
@parm uint32 | data | Writing data.
@rvalue SUCCESS | 
@rvalue FAILED | invalid parameter
@rvalue ERROR_PHY_INVALIDPHYNO | invalid PHY address
@rvalue ERROR_PHY_INVALIDREG | invalid PHY register
@rvalue ERROR_PHY_ACCESSBUSY | external PHY access busy
@comm
 */
int32 rtl8368s_setAsicExtPHYReg(uint32 phy, uint32 reg, uint32 data )
{
	int32 retVal;
	uint32 regData;

	if(phy > RTL8368S_PHY_EXT_ADDRESS_MAX)
		return ERROR_PHY_INVALIDPHYNO;

	if(reg > RTL8368S_PHY_REG_MAX)
		return ERROR_PHY_INVALIDREG;

	/*Check external PHY access busy indicator*/
	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_BUSY_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	if(regData & RTL8368S_PHY_EXT_BUSY_MASK)
		return ERROR_PHY_ACCESSBUSY;

	/*Write external PHY WR register*/
	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_WRDATA_REG, data);
	if (retVal !=  SUCCESS) 
		return retVal;	

	/*Write PHY access comand and address & register*/
	regData = RTL8368S_PHY_EXT_COMMAND_MASK |
				((phy<<RTL8368S_PHY_EXT_ADDR_OFFSET) & RTL8368S_PHY_EXT_ADDR_MASK ) | 
				((reg<<RTL8368S_PHY_EXT_REG_OFFSET) & RTL8368S_PHY_EXT_REG_MASK );
	

	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_CTRL_REG, regData);
	if (retVal !=  SUCCESS) 
		return retVal;

	return SUCCESS;	
}

/*
@func int32 | rtl8368s_getAsicExtPHYReg | get PHY registers .
@parm uint32 | phy | PHY address (0~31).
@parm uint32 | reg | PHY register (0~31).
@parm uint32* | data | Read data.
@rvalue SUCCESS | 
@rvalue FAILED | invalid parameter
@rvalue ERROR_PHY_INVALIDPHYNO | invalid PHY address
@rvalue ERROR_PHY_INVALIDREG | invalid PHY register
@rvalue ERROR_PHY_ACCESSBUSY | external PHY access busy
@comm
 */
int32 rtl8368s_getAsicExtPHYReg(uint32 phy, uint32 reg, uint32 *data )
{
	int32 retVal;
	uint32 regData,i;

	if(phy > RTL8368S_PHY_EXT_ADDRESS_MAX)
		return ERROR_PHY_INVALIDPHYNO;

	if(reg > RTL8368S_PHY_REG_MAX)
		return ERROR_PHY_INVALIDPHYPAGE;

	/*Write PHY access comand and address & register*/
	regData = RTL8368S_PHY_EXT_COMMAND_MASK |RTL8368S_PHY_EXT_RW_MASK | 
				((phy<<RTL8368S_PHY_EXT_ADDR_OFFSET) & RTL8368S_PHY_EXT_ADDR_MASK ) | 
				((reg<<RTL8368S_PHY_EXT_REG_OFFSET) & RTL8368S_PHY_EXT_REG_MASK );
	

	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_CTRL_REG, regData);
	if (retVal !=  SUCCESS) 
		return retVal;

	/*check PHY BUSY checking*/
	i=0;
	while(i<RTL8368S_TBL_CMD_CHECK_COUNTER)
	{
		retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_BUSY_REG,&regData);
		if(retVal !=  SUCCESS)
			return retVal;
		/*Check external PHY access busy indicator*/
		if(regData &RTL8368S_PHY_EXT_BUSY_MASK)
		{
			i++;
			if(i==RTL8368S_TBL_CMD_CHECK_COUNTER)
				return ERROR_PHY_ACCESSBUSY;
		}
		else
			break;
	}


	/*Get extern PHY RD data*/
	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_EXT_RDDATA_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;
	
	*data = regData;
		
	return SUCCESS;	
}


/*
@func int32 | rtl8368s_getAsicExtPHYReg_withPage | get PHY registers .
@parm uint32 | phy | PHY address (0~31).
@parm uint32 | page | page number(0~7).
@parm uint32 | reg | PHY register (0~31).
@parm uint32* | data | Read data.
@rvalue SUCCESS | 
@rvalue FAILED | invalid parameter
@rvalue ERROR_PHY_INVALIDPHYNO | invalid PHY address
@rvalue ERROR_PHY_INVALIDREG | invalid PHY register
@rvalue ERROR_PHY_ACCESSBUSY | external PHY access busy
@comm
 */
int32 rtl8368s_getAsicExtPHYReg_withPage(uint32 phy, uint32 page, uint32 reg, uint32 *data )
{
	int32 retVal;
	uint32 regData,i;

	if(phy > RTL8368S_PHY_EXT_ADDRESS_MAX)
		return ERROR_PHY_INVALIDPHYNO;

    if(page > RTL8368S_PHY_PAGE_MAX)
		return ERROR_PHY_INVALIDPHYPAGE;

	if(reg > RTL8368S_PHY_REG_MAX)
		return ERROR_PHY_INVALIDREG;

    /* Change page */
    retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_WRDATA_REG, page);
	if (retVal !=  SUCCESS) 
		return retVal;

    regData = RTL8368S_PHY_EXT_COMMAND_MASK | 
				((phy<<RTL8368S_PHY_EXT_ADDR_OFFSET) & RTL8368S_PHY_EXT_ADDR_MASK ) | 
				((RTL8368S_PHY_EXT_PAGE_REGNO<<RTL8368S_PHY_EXT_REG_OFFSET) & RTL8368S_PHY_EXT_REG_MASK );

    retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_CTRL_REG, regData);
	if (retVal !=  SUCCESS) 
		return retVal;

    /*check PHY BUSY checking*/
	i=0;
	while(i<RTL8368S_TBL_CMD_CHECK_COUNTER)
	{
		retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_BUSY_REG,&regData);
		if(retVal !=  SUCCESS)
			return retVal;
		/*Check external PHY access busy indicator*/
		if(regData &RTL8368S_PHY_EXT_BUSY_MASK)
		{
			i++;
			if(i==RTL8368S_TBL_CMD_CHECK_COUNTER)
				return ERROR_PHY_ACCESSBUSY;
		}
		else
			break;
	}

	/*Write PHY access comand and address & register*/
	regData = RTL8368S_PHY_EXT_COMMAND_MASK |RTL8368S_PHY_EXT_RW_MASK | 
				((phy<<RTL8368S_PHY_EXT_ADDR_OFFSET) & RTL8368S_PHY_EXT_ADDR_MASK ) | 
				((reg<<RTL8368S_PHY_EXT_REG_OFFSET) & RTL8368S_PHY_EXT_REG_MASK );
	

	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_EXT_CTRL_REG, regData);
	if (retVal !=  SUCCESS) 
		return retVal;

	/*check PHY BUSY checking*/
	i=0;
	while(i<RTL8368S_TBL_CMD_CHECK_COUNTER)
	{
		retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_BUSY_REG,&regData);
		if(retVal !=  SUCCESS)
			return retVal;
		/*Check external PHY access busy indicator*/
		if(regData &RTL8368S_PHY_EXT_BUSY_MASK)
		{
			i++;
			if(i==RTL8368S_TBL_CMD_CHECK_COUNTER)
				return ERROR_PHY_ACCESSBUSY;
		}
		else
			break;
	}


	/*Get extern PHY RD data*/
	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_EXT_RDDATA_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;
	
	*data = regData;
		
	return SUCCESS;	
}


/*=======================================================================
 *OTHERS
 *========================================================================*/
/*
@func int32 | rtl8368s_setAsicLedIndicateInfoConfig | Set Leds indicated information mode
@parm uint32 | ledNo | LED group number. There are 1 to 1 led mapping to each port in each led group. 
@parm enum RTL8368S_LEDCONF | config | Support 16 types configuration.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set LED indicated information configuration for each LED group with 1 to 1 led mapping to each port.
	Definition		LED Statuses			Description
	0000		LED_Off				LED pin Tri-State.
	0001		Dup/Col				Collision, Full duplex Indicator. Blinking every 43ms when collision happens. Low for full duplex, and high for half duplex mode.
	0010		Link/Act				Link, Activity Indicator. Low for link established. Link/Act Blinks every 43ms when the corresponding port is transmitting or receiving.
	0011		Spd1000				1000Mb/s Speed Indicator. Low for 1000Mb/s.
	0100		Spd100				100Mb/s Speed Indicator. Low for 100Mb/s.
	0101		Spd10				10Mb/s Speed Indicator. Low for 10Mb/s.
	0110		Spd1000/Act			1000Mb/s Speed/Activity Indicator. Low for 1000Mb/s. Blinks every 43ms when the corresponding port is transmitting or receiving.
	0111		Spd100/Act			100Mb/s Speed/Activity Indicator. Low for 100Mb/s. Blinks every 43ms when the corresponding port is transmitting or receiving.
	1000		Spd10/Act			10Mb/s Speed/Activity Indicator. Low for 10Mb/s. Blinks every 43ms when the corresponding port is transmitting or receiving.
	1001		Spd100 (10)/Act		10/100Mb/s Speed/Activity Indicator. Low for 10/100Mb/s. Blinks every 43ms when the corresponding port is transmitting or receiving.
	1010		Fiber				Fiber link Indicator. Low for Fiber.
	1011		Fault	Auto-negotiation 	Fault Indicator. Low for Fault.
	1100		Link/Rx				Link, Activity Indicator. Low for link established. Link/Rx Blinks every 43ms when the corresponding port is transmitting.
	1101		Link/Tx				Link, Activity Indicator. Low for link established. Link/Tx Blinks every 43ms when the corresponding port is receiving.
	1110		Master				Link on Master Indicator. Low for link Master established.
	1111		LED_Force			Force LED output, LED output value reference 
 */
int32 rtl8368s_setAsicLedIndicateInfoConfig(uint32 ledNo, enum RTL8368S_LEDCONF config)
{
	uint32 retVal, regValue;

	if(ledNo >=RTL8368S_LED_GROUP_MAX)
		return ERRNO_INPUT;

	if(config > LEDCONF_LEDFORCE)	
		return ERRNO_INPUT;


	/* Get register value */
	retVal = rtl8368s_getAsicReg(RTL8368S_LED_INDICATED_CONF_REG, &regValue); 	
	if (retVal !=  SUCCESS) 
		return retVal;

	regValue =  (regValue & (~(0xF<<(ledNo*4)))) | (config<<(ledNo*4));

	
	retVal = rtl8368s_setAsicReg(RTL8368S_LED_INDICATED_CONF_REG, regValue); 	

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicLedIndicateInfoConfig | Get Leds indicated information mode
@parm uint32 | ledNo | LED group number. There are 1 to 1 led mapping to each port in each led group. 
@parm enum RTL8368S_LEDCONF* | config | Support 16 types configuration.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can get LED indicated information configuration for each LED group.
 */
int32 rtl8368s_getAsicLedIndicateInfoConfig(uint32 ledNo, enum RTL8368S_LEDCONF* config)
{
	uint32 retVal, regValue;

	if(ledNo >=RTL8368S_LED_GROUP_MAX)
		return ERRNO_INPUT;

	/* Get register value */
	retVal = rtl8368s_getAsicReg(RTL8368S_LED_INDICATED_CONF_REG, &regValue); 	
	if (retVal !=  SUCCESS) 
		return retVal;
	
	*config = (regValue >> (ledNo*4)) & 0x000F;
		
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicForceLeds | Turn on/off Led of dedicated port
@parm uint32 | ledG0Msk | Turn on or turn off Leds of group 0, 1:on 0:off.
@parm uint32 | ledG1Msk | Turn on or turn off Leds of group 1, 1:on 0:off.
@parm uint32 | ledG2Msk | Turn on or turn off Leds of group 2, 1:on 0:off.
@parm uint32 | ledG3Msk | Turn on or turn off Leds of group 3, 1:on 0:off.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can turn on/off desired Led group of dedicated port while indicated information configuration of LED group is set to force mode.
 */
int32 rtl8368s_setAsicForceLeds(uint32 ledG0Msk, uint32 ledG1Msk, uint32 ledG2Msk, uint32 ledG3Msk)
{
	uint32 retVal, regValue;

	regValue = (ledG0Msk & RTL8368S_LED_0_FORCE_MASK) | ((ledG1Msk<<RTL8368S_LED_1_FORCE_OFF)&RTL8368S_LED_1_FORCE_MASK);

	retVal = rtl8368s_setAsicReg(RTL8368S_LED_0_1_FORCE_REG, regValue); 	
	if(retVal != SUCCESS)
		return retVal;

	regValue = (ledG2Msk & RTL8368S_LED_2_FORCE_MASK) | ((ledG3Msk<<RTL8368S_LED_3_FORCE_OFF)&RTL8368S_LED_3_FORCE_MASK);
	retVal = rtl8368s_setAsicReg(RTL8368S_LED_2_3_FORCE_REG, regValue); 	
	
	return retVal;
}

/*
@func int32 | rtl8368s_setAsicLedBlinkRate | Set led blinking rate ate mode 0 to mode 3
@parm enum RTL8368S_LEDBLINKRATE | blinkRate | Support 6 blink rates.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set LED blink rate at 43ms, 84ms, 120ms, 170ms, 340ms and 670ms.
 */
int32 rtl8368s_setAsicLedBlinkRate(enum RTL8368S_LEDBLINKRATE blinkRate)
{
	uint32 retVal;
	uint32 regData;

	if(blinkRate >=LEDBLINKRATE_MAX)
		return ERRNO_INPUT;

	regData = blinkRate << RTL8368S_LED_BLINKRATE_OFF;

	retVal = rtl8368s_setAsicRegBits(RTL8368S_LED_BLINK_REG,RTL8368S_LED_BLINKRATE_MSK,regData);

	return retVal;

}

/*
@func int32 | rtl8368s_getAsicLedBlinkRate | Get led blinking rate ate mode 0 to mode 3
@parm enum RTL8368S_LEDBLINKRATE* | blinkRate | Support 6 blink rates.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set LED blink rate at 43ms, 84ms, 120ms, 170ms, 340ms and 670ms.
 */
int32 rtl8368s_getAsicLedBlinkRate(enum RTL8368S_LEDBLINKRATE* blinkRate)
{
	uint32 retVal;
	uint32 regData;


	retVal = rtl8368s_getAsicReg(RTL8368S_LED_BLINK_REG,&regData);
	if(retVal !=  SUCCESS)
		return retVal;

	*blinkRate = (regData&RTL8368S_LED_BLINKRATE_MSK)>>RTL8368S_LED_BLINKRATE_OFF;

	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAiscPort4RGMIIControl | Set LED/Interrupt Control for P4 RGMII mode
@parm uint32 | enabled | 1:Enable LED/Interrupt for P4 RGMII, 0:Disable LED/Interrupt for P4 RGMII
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	Set LED/Interrupt Control for P4 RGMII mode
 */
int32 rtl8368s_setAiscPort4RGMIIControl(uint32 enabled)
{
    if((enabled != 0) && (enabled != 1))
        return FAILED;

    /* P4 RGMII LED Configuration */
    if(rtl8368s_setAsicRegBits(RTL8368S_INTERRUPT_CONTROL_REG, 
                               RTL8368S_INTERRUPT_P4_RGMII_LED_MSK, 
                               (enabled << RTL8368S_INTERRUPT_P4_RGMII_LED_OFF)) != SUCCESS)
        return FAILED;

    /* P4 RGMII Interrupt Configuration */
    if(rtl8368s_setAsicRegBits(RTL8368S_INTERRUPT_MASK_REG, 
                               RTL8368S_INTERRUPT_P4_FIBER_INT_MSK, 
                               (enabled << RTL8368S_INTERRUPT_P4_FIBER_INT_OFF) ) != SUCCESS)
        return FAILED;

    if(rtl8368s_setAsicRegBits(RTL8368S_INTERRUPT_MASK_REG, 
                               RTL8368S_INTERRUPT_P4_UTP_INT_MSK, 
                               (enabled << RTL8368S_INTERRUPT_P4_UTP_INT_OFF) ) != SUCCESS)
        return FAILED;

    return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicMacForceLink | Set mac force linking configuration.
@parm enum PORTID | port | Port/MAC number (0~5).
@parm enum MACLINKMODE | force | Mac link mode 1:force mode 0:normal
@parm enum PORTLINKSPEED | speed | Speed of the port 0b00-10M, 0b01-100M,0b10-1000M, 0b11 reserved.
@parm enum PORTLINKDUPLEXMODE | duplex | Deuplex mode 0b0-half duplex 0b1-full duplex.
@parm uint32 | link | Link status 0b0-link down b1-link up.
@parm uint32 | txPause | Pause frame transmit ability of the port 0b0-inactive 0b1-active.
@parm uint32 | rxPause | Pause frame response ability of the port 0b0-inactive 0b1-active.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
  	This API can set Port/MAC force mode properties. 
 */
int32 rtl8368s_setAsicMacForceLink(enum PORTID port,enum MACLINKMODE force,enum PORTLINKSPEED speed,enum PORTLINKDUPLEXMODE duplex,uint32 link,uint32 txPause,uint32 rxPause)
{
	uint32 retVal;
	uint32 macData;
	uint32 regBits;
	uint32 regAddr;
	uint32 regData;
	
	/* Invalid input parameter */
	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	/*not force mode*/
	if(MAC_NORMAL == force)
	{
		retVal = rtl8368s_getAsicReg(RTL8368S_MAC_FORCE_CTRL_REG,&regData);
		if (retVal !=  SUCCESS) 
			return retVal;
		
		regData = regData & (~(1<<port));

		retVal = rtl8368s_setAsicReg(RTL8368S_MAC_FORCE_CTRL_REG,regData);
		if (retVal !=  SUCCESS) 
			return retVal;

		return SUCCESS;
	}


	if(speed > SPD_1000M)
		return ERROR_MAC_INVALIDSPEED;

	/*prepare force status first*/
	macData = speed;

	if(duplex)
	{
		macData = macData | (duplex<<2);
	}

	if(link)
	{
		macData = macData | (link<<4);
	}

	if(txPause)
	{
		macData = macData | (txPause<<5);
	}
	
	if(rxPause)
	{
		macData = macData | (rxPause<<6);
	}
	
	regBits = 0xFF << (8*(port&0x01));
	macData = macData <<(8*(port&0x01));
	
	/* Set register value */
	regAddr = RTL8368S_PORT_ABILITY_BASE + (port>>1);


	retVal= rtl8368s_setAsicRegBits(regAddr,regBits,macData);
	if (retVal !=  SUCCESS) 
		return retVal;


	/* Set register value */
	retVal = rtl8368s_getAsicReg(RTL8368S_MAC_FORCE_CTRL_REG,&regData);
	if (retVal !=  SUCCESS) 
		return retVal;

	regData = regData | (1<<port);

	retVal = rtl8368s_setAsicReg(RTL8368S_MAC_FORCE_CTRL_REG,regData);
	if (retVal !=  SUCCESS) 
		return retVal;

	return SUCCESS;
}


/*
@func int32 | rtl8368s_setAsicCableTesting | Set port cable testing.
@parm enum PORTID | port | Physical port number (0~4).
@parm uint32 | enabled | Per-Port cable testing function 1: enabled, 0: disabled
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can start realtek cable testing function.
 */
int32 rtl8368s_setAsicCableTesting( enum PORTID port, uint32 enabled)
{

	int32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;
	
	retVal = rtl8368s_setAsicRegBit(RTL8368S_CABLE_TESTING_CTRL_REG, (RTL8368S_CABLE_TESTING_START_OFF + port), enabled);

	return retVal;
}


/*
@func int32 | rtl8368s_getAsicRtctChannelStatus | Get channel cable testing result of each port.
@parm enum PORTID | port | Physical port number (0~4).
@parm enum CHANNELID | channel | Channel id of each port (0~3).
@parm uint32 | *impmis | Channel impedance mismatch status 1: impedance mismatch, 0: normal
@parm uint32 | *open | Channel open status 1: open, 0: normal
@parm uint32 | *short | Channel short status  1: short, 0: normal
@parm uint32 | *length | Cable length of channel (unit: 2.5cm).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can Get realtek cable testing function result.
 */
int32 rtl8368s_getAsicRtctChannelStatus(enum PORTID port, enum CHANNELID channel,uint32* impmis,uint32* open_state,uint32* short_state,uint32* length)
{
	int32 retVal;
	uint32 regData;
	
	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;
	if(channel >=CH_MAX)
		return ERRNO_INPUT;

    /* Status checking, if RTCT function is not finished, return FAILED. */
    retVal = rtl8368s_getAsicReg(RTL8368S_CABLE_TESTING_ACT_REG, &regData);
	if(retVal !=  SUCCESS)
		return retVal;

    if(regData & (0x1 << (RTL8368S_CABLE_TESTING_ACT_OFF + port)))
        return FAILED;

    /* RTCT function finished, Start to get information */
    retVal = rtl8368s_getAsicReg((RTL8368S_CABLE_TESTING_INFO_BASE + (port * 5)), &regData);
	if(retVal !=  SUCCESS)
		return retVal;

    /* Mismatch */
    if(regData & RTL8368S_CABLE_TESTING_CH_MIS_MSK)
		*impmis = 1;
	else
		*impmis = 0;

    /* Open */
    if(regData & RTL8368S_CABLE_TESTING_CH_OPEN_MSK)
		*open_state = 1;
	else
		*open_state = 0;

    /* Short */
    if(regData & RTL8368S_CABLE_TESTING_CH_SHORT_MSK)
		*short_state = 1;
	else
		*short_state = 0;

    /* Length */
    retVal = rtl8368s_getAsicReg((RTL8368S_CABLE_TESTING_INFO_BASE + (port * 5) + channel + 1), &regData);
	if(retVal !=  SUCCESS)
		return retVal;
	
	*length  = regData;
    return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicAgeTimerSpeed | Set LUT agging out speed
@parm uint32 | timer | Agging out timer 0:Has been aged out.
@parm uint32 | speed | Agging out speed 0-fastest 3-slowest.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid parameter.
@comm
 	The API can set LUT agging out period for each entry. Differet {timer, speed} parameter will make asic do agging out task
 	at different period time. Following description is times period for different timer,speed combination setting.
			Timer		1		2		3		4		5		6		7
	Speed
	0					14.3s	28.6s	42.9s	57.2s	1.19m	1.43m	1.67m
	1					28.6s	57.2s	1.43m	1.9m	2.38m	2.86m	3.34m
	2					57.2s	1.9m	2.86m	3.81m	4.77m	5.72m	6.68m
	3					1.9m	3.8m	5.72m	7.63m	9.54m	11.45m	13.36m
 	(s:Second m:Minute)
 */
int32 rtl8368s_setAsicAgeTimerSpeed( uint32 timer, uint32 speed)
{
	uint32 retVal;
	uint32 regData;
	uint32 regBits;

	if(timer>RTL8368S_AGE_TIMER_MAX)
		return ERRNO_INPUT;

	if(speed >RTL8368S_AGE_SPEED_MAX)
		return ERRNO_INPUT;
	
	regBits = RTL8368S_AGE_TIMER_MSK | RTL8368S_AGE_SPEED_MSK;

	regData = ((timer<<RTL8368S_AGE_TIMER_OFF) & RTL8368S_AGE_TIMER_MSK) | ((speed<<RTL8368S_AGE_SPEED_OFF) & RTL8368S_AGE_SPEED_MSK);

	retVal = rtl8368s_setAsicRegBits(RTL8368S_SECURITY_CTRL2_REG,regBits,regData);		
	
	return retVal;	
}

/*
@func int32 | rtl8368s_getAsicAgeTimerSpeed | Set LUT agging out speed
@parm uint32* | timer | Agging out timer 0:Has been aged out.
@parm uint32* | speed | Agging out speed 0-fastest 3-slowest.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid parameter.
@comm
 	The API can set LUT agging out period for each entry. Differet {timer, speed} parameter will make asic do agging out task
 	at different period time. Following description is times period for different timer,speed combination setting.
			Timer		1		2		3		4		5		6		7
	Speed
	0					14.3s	28.6s	42.9s	57.2s	1.19m	1.43m	1.67m
	1					28.6s	57.2s	1.43m	1.9m	2.38m	2.86m	3.34m
	2					57.2s	1.9m	2.86m	3.81m	4.77m	5.72m	6.68m
	3					1.9m	3.8m	5.72m	7.63m	9.54m	11.45m	13.36m
 	(s:Second m:Minute)
 */
int32 rtl8368s_getAsicAgeTimerSpeed( uint32* timer, uint32* speed)
{
	uint32 regData;
	int32 retVal;

	retVal = rtl8368s_getAsicReg(RTL8368S_SECURITY_CTRL2_REG,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*timer =  (regData & RTL8368S_AGE_TIMER_MSK) >> RTL8368S_AGE_TIMER_OFF;
	
	*speed =  (regData & RTL8368S_AGE_SPEED_MSK) >> RTL8368S_AGE_SPEED_OFF;

	return SUCCESS;

}


/*
@func int32 | rtl8368s_setAsicPortIsolation | Set port isolation function enable usage configuration.
@parm enum PORTID | port | Egress port number (0~8).
@parm uint32 | enabled | Isolation function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
	This API can set ASIC port isolation functions' uage. It is enabled per port. By using isolation setting, 
	user can isolate physical ports' group without VLAN configuration. This function will physically seperate networking 
	flow and isolate network within desired ports only.
 */

int32 rtl8368s_setAsicPortIsolation(enum PORTID port, uint32 enable)
{
	int32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;
	retVal = rtl8368s_setAsicRegBit(RTL8368S_PORT_ISOLATION_BASE+(port),RTL8368S_PORT_ISOLATION_ENABLE_OFF, enable);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicPortIsolation | Get port isolation function enable usage configuration.
@parm enum PORTID | port | Egress port number (0~8).
@parm uint32* | enabled | Isolation function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
	This API can get ASIC port isolation functions' uage. By using isolation setting, user can isolate physical
	ports' group without VLAN configuration. This function will physically seperate networking flow and isolate network 
	within desired ports only.
 */

int32 rtl8368s_getAsicPortIsolation( enum PORTID port, uint32* enabled)
{
	uint32 regData;
	int32 retVal;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;

	retVal = rtl8368s_getAsicReg(RTL8368S_PORT_ISOLATION_BASE+(port),&regData);
	if(retVal != SUCCESS)
		return retVal;

	if(regData & RTL8368S_PORT_ISOLATION_ENABLE_MSK)
		*enabled = 1;
	else
		*enabled = 0;

	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicPortIsolationConfig | Set ASIC port isolation control configuration. 
@parm enum PORTID | port | Source port number (0~8).
@uint32 | portmask | Output port mask (0-8).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can isolation source port output to output ports represented by port mask. When the API is used,
	packets from source port will not deliver to other ports that are not in port mask. Other ports not in port mask 
	can still deliver packets to source port. By using isolation setting, user can isolate physical ports' group without
	VLAN configuration. This function will physically seperate networking flow and isolate network within desired ports only.
*/
int32 rtl8368s_setAsicPortIsolationConfig(enum PORTID port, uint32 portmask)
{

	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;
 
 	retVal = rtl8368s_getAsicReg(RTL8368S_PORT_ISOLATION_BASE+(port),&regData);
	regData = (regData & RTL8368S_PORT_ISOLATION_ENABLE_MSK)|((portmask<<RTL8368S_PORT_ISOLATION_PORTS_OFF)&RTL8368S_PORT_ISOLATION_PORTS_MSK);
	retVal = rtl8368s_setAsicReg(RTL8368S_PORT_ISOLATION_BASE+(port),regData);

	return retVal;
}
/*
@func int32 | rtl8368s_getAsicPortIsolationConfig | Get ASIC port isolation control configuration. 
@parm enum PORTID | port | Source port number (0~8).
@uint32* | portmask | Output port mask (0-8).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
 	The API can be used to get the port isolation configuration.
*/

int32 rtl8368s_getAsicPortIsolationConfig(enum PORTID port, uint32* portmask)
{

	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;
 
	retVal = rtl8368s_getAsicReg(RTL8368S_PORT_ISOLATION_BASE+port, &regData);

	*portmask = (regData & RTL8368S_PORT_ISOLATION_PORTS_MSK) >>RTL8368S_PORT_ISOLATION_PORTS_OFF;

	return retVal;
}

/*
@func int32 | rtl8368s_setAsicSpecialCongestModeConfig | Set ASIC special congest mode configuration. 
@parm enum PORTID | port | port number (0~8).
@uint32 | sustain | sustain timer (0-15).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can set ASIC special congest mode per port. It is a timer that if the source port was blocked by 
	pause frame and get into congestion state, the port will return normal state after sustain time.
*/

int32 rtl8368s_setAsicSpecialCongestModeConfig(enum PORTID port, uint32 sustain)
{

	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;

	 retVal = rtl8368s_getAsicReg(RTL8368S_CONGEST_MODE_CONTROL_BASE+(port*RTL8368S_CONGEST_MODE_CONTROL_OFF),&regData);
	 
	regData = (regData&0xFFF0)|((sustain<<RTL8368S_CONGEST_MODE_SUSTAIN_OFF)&RTL8368S_CONGEST_MODE_SUSTAIN_MSK);
	
	retVal = rtl8368s_setAsicReg(RTL8368S_CONGEST_MODE_CONTROL_BASE+(port*RTL8368S_CONGEST_MODE_CONTROL_OFF),regData);

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSpecialCongestModeConfig | Get ASIC special congest mode configuration. 
@parm enum PORTID | port | port number (0~8).
@uint32* | sustain | sustain timer (0-15).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can get ASIC special congest mode setup per port.
*/
int32 rtl8368s_getAsicSpecialCongestModeConfig(enum PORTID port, uint32* sustain)
{

	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;
 
	retVal = rtl8368s_getAsicReg(RTL8368S_CONGEST_MODE_CONTROL_BASE+(port*RTL8368S_CONGEST_MODE_CONTROL_OFF),&regData);

	*sustain = (regData&RTL8368S_CONGEST_MODE_SUSTAIN_MSK)>>RTL8368S_CONGEST_MODE_SUSTAIN_OFF;

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicSpecialCongestModeTimer | Get ASIC special congest mode timer. 
@parm enum PORTID | port | port number (0~8).
@uint32* | timer | time (0-15).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
	The API can get ASIC special congest mode time from congestion state to normal state now per port.
*/
int32 rtl8368s_getAsicSpecialCongestModeTimer(enum PORTID port, uint32* timer)
{

	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_INPUT;
 
	retVal = rtl8368s_getAsicReg(RTL8368S_CONGEST_MODE_CONTROL_BASE+(port*RTL8368S_CONGEST_MODE_CONTROL_OFF),&regData);

	*timer = (regData&RTL8368S_CONGEST_MODE_TIMER_MSK)>>RTL8368S_CONGEST_MODE_TIMER_OFF;

	return retVal;
}

/*
@func int32 | rtl8368s_setAsicOamParser | Set OAM parser state
@parm enum PORTID | port | Physical port number (0~7).
@parm enum OAMPARACT | parser | Per-Port OAM parser state
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PORT_NUM | Invalid port number.
@comm
	The API can set per-port OAM parser state. Using OAM parser and multiplex configuration to implement 
	OAM loopback testing.
 */
int32 rtl8368s_setAsicOamParser(enum PORTID port, enum OAMPARACT parser)
{
	int32 retVal;
	uint32 regData;
	uint32 regBits;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	if(parser > OAM_PARFWDCPU)
		return ERRNO_OAM_PARACT;
	
	regBits = 0x3 << (2*port);
	regData = parser <<(2*port);

	retVal = rtl8368s_setAsicRegBits(RTL8368S_OAM_PARSER_REG,regBits,regData);		

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicOamParser | Set OAM parser state
@parm enum PORTID | port | Physical port number (0~7).
@parm enum OAMPARACT* | parser | Per-Port OAM parser state
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PORT_NUM | Invalid port number.
@comm
	The API can get per-port OAM parser state. Using OAM parser and multiplex configuration to implement 
	OAM loopback testing.
 */
int32 rtl8368s_getAsicOamParser(enum PORTID port, enum OAMPARACT* parser)
{
	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	retVal = rtl8368s_getAsicReg(RTL8368S_OAM_PARSER_REG,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*parser =  (regData & (0x3 << (2*port))) >> (2*port);
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicOamMultiplexer | Set OAM multiplexer state
@parm enum PORTID | port | Physical port number (0~7).
@parm enum OAMMULACT | multiplexer | Per-Port OAM multiplexer state
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PORT_NUM | Invalid port number.
@comm
	The API can set per-port OAM multiplexer state. Using OAM parser and multiplex configuration to implement 
	OAM loopback testing.
 */
int32 rtl8368s_setAsicOamMultiplexer(enum PORTID port, enum OAMMULACT multiplexer)
{
	int32 retVal;
	uint32 regData;
	uint32 regBits;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	if(multiplexer > OAM_MULCPU)
		return ERRNO_OAM_MULACT;
	
	regBits = 0x3 << (2*port);
	regData = multiplexer <<(2*port);

	retVal = rtl8368s_setAsicRegBits(RTL8368S_OAM_MULTIPLEXER_REG,regBits,regData);		

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicOamMultiplexer | Get OAM multiplexer state
@parm enum PORTID | port | Physical port number (0~7).
@parm enum OAMMULACT* | multiplexer | Per-Port OAM multiplexer state
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_PORT_NUM | Invalid port number.
@comm
	The API can get per-port OAM multiplexer state. Using OAM parser and multiplex configuration to implement 
	OAM loopback testing.
 */
int32 rtl8368s_getAsicOamMultiplexer(enum PORTID port, enum OAMMULACT* multiplexer)
{
	int32 retVal;
	uint32 regData;

	if(port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	retVal = rtl8368s_getAsicReg(RTL8368S_OAM_MULTIPLEXER_REG,&regData);
	if(retVal != SUCCESS)
		return retVal;

	*multiplexer =  (regData & (0x3 << (2*port))) >> (2*port);
	
	return SUCCESS;
}

/*
@func int32 | rtl8368s_setAsicMacAddress | Set Mac Address of Switch
@parm ether_addr_t | macAsic | MAC
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can set MAC address of switch. The MAC address will be attached to SA of PAUSE frame.
 */
int32 rtl8368s_setAsicMacAddress(ether_addr_t macAsic)
{
	int32 retVal;
	uint32 regData;
	uint16 *accessPtr;
	uint32 i;


	accessPtr =  (uint16*)&macAsic;

	regData = *accessPtr;
	for(i=0;i<=2;i++)
	{
		retVal = rtl8368s_setAsicReg(RTL8368S_SWITCH_MAC_BASE+i,regData);
		if(retVal !=  SUCCESS)
			return retVal;

		accessPtr ++;
		regData = *accessPtr;
	}

	return retVal;
}

/*
@func int32 | rtl8368s_getAsicMacAddress | Set Mac Address of Switch
@parm ether_addr_t* | macAsic | MAC
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	The API can get MAC address of switch. The MAC address will be attached to SA of PAUSE frame.
 */
int32 rtl8368s_getAsicMacAddress(ether_addr_t* macAsic)
{
	int32 retVal;
	uint32 regData;
	uint16 *accessPtr;
	uint32 i;


	accessPtr =  (uint16*)macAsic;

	for(i=0;i<=2;i++)
	{
		retVal = rtl8368s_getAsicReg(RTL8368S_SWITCH_MAC_BASE+i,&regData);
		if(retVal !=  SUCCESS)
			return retVal;

		*accessPtr = regData;

		accessPtr ++;
	}

	return retVal;
}


/*
@func int32 | rtl8368s_setAsicGreenFeature | Set green ethernet function.
@parm uint32 | enable | Green ethernet function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
	This API can set ASIC green ethernet functions' uage. When green feature function was enabled, RTL8366S/SR will dynamic estimate
connection cable length and turn on different Tx/Rx power mode with cable estimate result.
 */
int32 rtl8368s_setAsicGreenFeature(uint32 enable)
{
	uint32 regData;

	if(enable>1)
		return FAILED;

	if(SUCCESS != rtl8368s_getAsicReg(RTL8368S_REVISION_ID_REG,&regData))
		return FAILED;

	if (enable==1)
	{
		if (regData==0x0001)
		{
			if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_GREEN_FEATURE_REG, 0x0007))
				return FAILED;
		}
		else
		{
			if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_GREEN_FEATURE_REG, 0x0003))
				return FAILED;
		}
	}
	else
	{
		if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_GREEN_FEATURE_REG, 0))
			return FAILED;
	}

	return SUCCESS;
}

/*
@func int32 | rtl8368s_getAsicGreenFeature | get green ethernet function.
@parm uint32* | enable | Green ethernet function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
	This API can get ASIC green ethernet functions' uage. When green feature function was enabled, RTL8366S/SR will dynamic estimate
connection cable length and turn on different Tx/Rx power mode with cable estimate result.
 */
int32 rtl8368s_getAsicGreenFeature(uint32* enable)
{
	int32 retVal;
	int32 regValue;
	uint32 regData;

	retVal = rtl8368s_getAsicReg(RTL8368S_GREEN_FEATURE_REG,&regValue);

	if(SUCCESS != retVal)
		return retVal;
	
	if(SUCCESS != rtl8368s_getAsicReg(RTL8368S_REVISION_ID_REG,&regData))
		return FAILED;

	
	if (regData==0x0001)
	{
		if((regValue & RTL8368S_GREEN_FEATURE_MSK)==RTL8368S_GREEN_FEATURE_MSK)
		{
			*enable = 1;	
		}
		else
		{
			*enable = 0;	
		}
	}
	else
	{
		if((regValue & RTL8368S_GREEN_FEATURE_MSK)==0x0003)
		{
			*enable = 1;	
		}
		else
		{
			*enable = 0;	
		}
	}
	return SUCCESS;
}


/*
@func int32 | rtl8368s_setAsicPowerSaving | Set power saving function.
@parm uint32 | phyNo | PHY number (0~3).
@parm uint32 | enabled | Qos function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
  	This API can set ASIC power saving functions' uage for dedicated PHY.When UTP is not plugged-in, this port only transmits link 
  pulse on TX and detects signal on RX. When UTP is plugged-in in power saving mode, this port should link on proper status by Nway 
  or parallel detection result.
 */
int32 rtl8368s_setAsicPowerSaving(uint32 phyNo,uint32 enabled)
{
	uint32 retVal;
	uint32 phySmiAddr;
	uint32 phyData;
	
	if(phyNo > RTL8368S_PHY_NO_MAX)
		return ERROR_PHY_INVALIDPHYNO;


	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_READ);
	if (retVal !=  SUCCESS) 
		return retVal;

	phySmiAddr = 0x8000 | (1<<(phyNo +RTL8368S_PHY_NO_OFFSET)) | 
			((RTL8368S_POWER_SAVING_PAGE<<RTL8368S_PHY_PAGE_OFFSET)&RTL8368S_PHY_PAGE_MASK) | 
			(RTL8368S_POWER_SAVING_REG&RTL8368S_PHY_REG_MASK);
	
	retVal = rtl8368s_setAsicReg(phySmiAddr, 0);
	if (retVal !=  SUCCESS) 
		return retVal;


	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_DATA_REG,&phyData);
	if(retVal !=  SUCCESS)
		return retVal;


	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_WRITE);
	if (retVal !=  SUCCESS) 
		return retVal;


	if(enabled)
	{
		phyData = phyData | RTL8368S_POWER_SAVING_BIT_MSK;
	}	
	else
	{
		phyData = phyData & (~RTL8368S_POWER_SAVING_BIT_MSK);
	}	
	
		
	retVal = rtl8368s_setAsicReg(phySmiAddr, phyData);
	if (retVal !=  SUCCESS) 
		return retVal;

	return SUCCESS;

}


/*
@func int32 | rtl8368s_getAsicPowerSaving | Set power saving function.
@parm uint32 | phyNo | PHY number (0~3).
@parm uint32* | enabled | Qos function usage 1:enable, 0:disabled.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
  	This API can get ASIC power saving functions' uage for dedicated PHY.When UTP is not plugged-in, this port only transmits link 
  pulse on TX and detects signal on RX. When UTP is plugged-in in power saving mode, this port should link on proper status by Nway 
  or parallel detection result.
 */
int32 rtl8368s_getAsicPowerSaving(uint32 phyNo,uint32* enabled)
{
	uint32 retVal;
	uint32 phySmiAddr;
	uint32 phyData;
	
	if(phyNo > RTL8368S_PHY_NO_MAX)
		return ERROR_PHY_INVALIDPHYNO;


	retVal = rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_READ);
	if (retVal !=  SUCCESS) 
		return retVal;

	phySmiAddr = 0x8000 | (1<<(phyNo +RTL8368S_PHY_NO_OFFSET)) | 
			((RTL8368S_POWER_SAVING_PAGE<<RTL8368S_PHY_PAGE_OFFSET)&RTL8368S_PHY_PAGE_MASK) | 
			(RTL8368S_POWER_SAVING_REG&RTL8368S_PHY_REG_MASK);
	
	retVal = rtl8368s_setAsicReg(phySmiAddr, 0);
	if (retVal !=  SUCCESS) 
		return retVal;


	retVal = rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_DATA_REG,&phyData);
	if(retVal !=  SUCCESS)
		return retVal;



	if(phyData & RTL8368S_POWER_SAVING_BIT_MSK) 
	{
		*enabled = 1;
	}	
	else
	{
		*enabled = 0;
	}	
	
	return SUCCESS;
}


//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//
//	copy from "rtl8368s_asicdrv.c"
//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

/*
* Copyright c                  Realtek Semiconductor Corporation, 2008 
* All rights reserved.
* 
* Program : RTL8366RB switch high-level API
* Abstract : 
* Author : H.M Chung (hmchung@realtek.com.tw), Kobe Wu (kobe_wu@realtek.com.tw)                 
* $Id: rtl8366rb_api.c,v 1.18 2008-06-25 06:22:44 kobe_wu Exp $
*/
/*	@doc rtl8366rb_api

	@module rtl8366rb_api.c - RTL8366RB switch high-level API documentation	|
	This document explains high-level API interface of the asic. 
	@normal 

	Copyright <cp>2008 Realtek<tm> Semiconductor Cooperation, All Rights Reserved.

 	@head3 List of Symbols |
 	Here is a list of all functions and variables in this module.

 	@index | rtl8366rb_api
*/

#ifndef	CONFIG_BUFFALO
#include "rtl8366rb_api.h"
#include "rtl8366rb_api_ext.h"
#include "rtl8368s_errno.h"
#include "rtl8368s_asicdrv.h"
#include "rtl8368s_reg.h"

#else	//CONFIG_BUFFALO
#include "rtl8366rb_api.h"
#include "rtl8366rb_phy.h"

#endif	//CONFIG_BUFFALO
  
#define DELAY_800MS_FOR_CHIP_STATABLE() {  }






static void _rtl8366rb_copy(uint8 *dest, const uint8 *src, uint32 n)
{
	uint32 i;
	
	for(i = 0; i < n; i++)
		dest[i] = src[i];
}

static int32 _rtl8366rb_cmp(const uint8 *s1, const uint8 *s2, uint32 n)
{
	uint32 i;
	
	for(i = 0; i < n; i++)
	{
		if(s1[i] != s2[i])
			return FAILED;/* not equal */
	}

	return SUCCESS;
}

static void _rtl8366rb_powof2(uint32 val, uint32 *ret)
{
	*ret = 1;
	
	for(; val > 0; val--)
		*ret = (*ret)*2;
}


/*
@func int32 | rtl8366rb_setEthernetPHY | Set ethernet PHY registers for desired ability.
@parm uint32 | phy | PHY number (0~4).
@parm phyAbility_t | ability | Ability structure
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	If Full_1000 bit is set to 1, the AutoNegotiation will be automatic 
    set to 1. While both AutoNegotiation and Full_1000 are set to 0, the 
    PHY speed and duplex selection will be set as following 100F > 100H 
    > 10F > 10H priority sequence.
*/
int32 rtl8366rb_setEthernetPHY(uint32 phy, rtl8366rb_phyAbility_t *ptr_ability)
{
	uint32 phyData;
	
	uint32 phyEnMsk0;
	uint32 phyEnMsk4;
	uint32 phyEnMsk9;
	
	if(phy > RTL8366RB_PHY_NO_MAX)
		return FAILED;

	phyEnMsk0 = 0;
	phyEnMsk4 = 0;
	phyEnMsk9 = 0;


	if(1 == ptr_ability->Half_10)
	{
		/*10BASE-TX half duplex capable in reg 4.5*/
		phyEnMsk4 = phyEnMsk4 | (1<<5);

		/*Speed selection [1:0] */
		/* 11=Reserved*/
		/* 10= 1000Mpbs*/
		/* 01= 100Mpbs*/
		/* 00= 10Mpbs*/		
		phyEnMsk0 = phyEnMsk0 & (~(1<<6));
		phyEnMsk0 = phyEnMsk0 & (~(1<<13));
	}

	if(1 == ptr_ability->Full_10)
	{
		/*10BASE-TX full duplex capable in reg 4.6*/
		phyEnMsk4 = phyEnMsk4 | (1<<6);
		/*Speed selection [1:0] */
		/* 11=Reserved*/
		/* 10= 1000Mpbs*/
		/* 01= 100Mpbs*/
		/* 00= 10Mpbs*/		
		phyEnMsk0 = phyEnMsk0 & (~(1<<6));
		phyEnMsk0 = phyEnMsk0 & (~(1<<13));

		/*Full duplex mode in reg 0.8*/
		phyEnMsk0 = phyEnMsk0 | (1<<8);
		
	}

	if(1 == ptr_ability->Half_100)
	{
		//PRINT("if(1 == ptr_ability->Half_100)\n ");
		/*100BASE-TX half duplex capable in reg 4.7*/
		phyEnMsk4 = phyEnMsk4 | (1<<7);
		/*Speed selection [1:0] */
		/* 11=Reserved*/
		/* 10= 1000Mpbs*/
		/* 01= 100Mpbs*/
		/* 00= 10Mpbs*/		
		phyEnMsk0 = phyEnMsk0 & (~(1<<6));
		phyEnMsk0 = phyEnMsk0 | (1<<13);
	}


	if(1 == ptr_ability->Full_100)
	{
		/*100BASE-TX full duplex capable in reg 4.8*/
		phyEnMsk4 = phyEnMsk4 | (1<<8);
		/*Speed selection [1:0] */
		/* 11=Reserved*/
		/* 10= 1000Mpbs*/
		/* 01= 100Mpbs*/
		/* 00= 10Mpbs*/		
		phyEnMsk0 = phyEnMsk0 & (~(1<<6));
		phyEnMsk0 = phyEnMsk0 | (1<<13);
		/*Full duplex mode in reg 0.8*/
		phyEnMsk0 = phyEnMsk0 | (1<<8);
	}
	
	
	if(1 == ptr_ability->Full_1000)
	{
		/*1000 BASE-T FULL duplex capable setting in reg 9.9*/
		phyEnMsk9 = phyEnMsk9 | (1<<9);

		/*Speed selection [1:0] */
		/* 11=Reserved*/
		/* 10= 1000Mpbs*/
		/* 01= 100Mpbs*/
		/* 00= 10Mpbs*/		
		phyEnMsk0 = phyEnMsk0 | (1<<6);
		phyEnMsk0 = phyEnMsk0 & (~(1<<13));
	

		/*Auto-Negotiation setting in reg 0.12*/
#if 0
		phyEnMsk0 = phyEnMsk0 | (1<<12);
#else
		if(ptr_ability->AutoNegotiation != 1)
			return FAILED;			
#endif
	}
	
	if(1 == ptr_ability->AutoNegotiation)
	{
		/*Auto-Negotiation setting in reg 0.12*/
		phyEnMsk0 = phyEnMsk0 | (1<<12);
	}

	if(1 == ptr_ability->AsyFC)
	{
		/*Asymetric flow control in reg 4.11*/
		phyEnMsk4 = phyEnMsk4 | (1<<11);
	}
	if(1 == ptr_ability->FC)
	{
		/*Flow control in reg 4.10*/
		phyEnMsk4 = phyEnMsk4 | (1<<10);
	}

	
	/*1000 BASE-T control register setting*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_1000_BASET_CONTROL_REG,&phyData))
		return FAILED;

	phyData = (phyData & (~0x0200)) | phyEnMsk9 ;
		
	if(SUCCESS != rtl8368s_setAsicPHYRegs(phy,0,RTL8366RB_PHY_1000_BASET_CONTROL_REG,phyData))
		return FAILED;

	/*Auto-Negotiation control register setting*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_AN_ADVERTISEMENT_REG,&phyData))
		return FAILED;

	phyData = (phyData & (~0x0DE0)) | phyEnMsk4;
		
	if(SUCCESS != rtl8368s_setAsicPHYRegs(phy,0,RTL8366RB_PHY_AN_ADVERTISEMENT_REG,phyData))
		return FAILED;


	/*Control register setting and restart auto*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_CONTROL_REG,&phyData))
		return FAILED;


	phyData = (phyData & (~0x3140)) | phyEnMsk0;
	/*If have auto-negotiation capable, then restart auto negotiation*/
	if(1 == ptr_ability->AutoNegotiation)
	{
		phyData = phyData | (1 << 9);
	}
	
	if(SUCCESS != rtl8368s_setAsicPHYRegs(phy,0,RTL8366RB_PHY_CONTROL_REG,phyData))
		return FAILED;

	
	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getEthernetPHY | Get PHY ability through PHY registers.
@parm uint32 | phy | PHY number (0~4).
@parm phyAbility_t* | ptr_ability | Ability structure
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	Get the capablity of specified PHY.
*/
int32 rtl8366rb_getEthernetPHY(uint32 phy, rtl8366rb_phyAbility_t *ptr_ability)
{
	uint32 phyData0;
	uint32 phyData4;
	uint32 phyData9;
	

	if(phy > RTL8366RB_PHY_NO_MAX)
		return FAILED;


	/*Control register setting and restart auto*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_CONTROL_REG,&phyData0))
		return FAILED;

	/*Auto-Negotiation control register setting*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_AN_ADVERTISEMENT_REG,&phyData4))
		return FAILED;

	/*1000 BASE-T control register setting*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_1000_BASET_CONTROL_REG,&phyData9))
		return FAILED;

	if(phyData9 & (1<<9))
		ptr_ability->Full_1000 = 1;
	else
		ptr_ability->Full_1000 = 0;

	if(phyData4 & (1<<11))
		ptr_ability->AsyFC = 1;
	else
		ptr_ability->AsyFC = 0;

	if(phyData4 & (1<<10))
		ptr_ability->FC = 1;
	else
		ptr_ability->FC = 0;
	

	if(phyData4 & (1<<8))
		ptr_ability->Full_100= 1;
	else
		ptr_ability->Full_100= 0;
	
	if(phyData4 & (1<<7))
		ptr_ability->Half_100= 1;
	else
		ptr_ability->Half_100= 0;

	if(phyData4 & (1<<6))
		ptr_ability->Full_10= 1;
	else
		ptr_ability->Full_10= 0;
	
	if(phyData4 & (1<<5))
		ptr_ability->Half_10= 1;
	else
		ptr_ability->Half_10= 0;


	if(phyData0 & (1<<12))
		ptr_ability->AutoNegotiation= 1;
	else
		ptr_ability->AutoNegotiation= 0;

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getPHYLinkStatus | Get ethernet PHY linking status
@parm uint32 | phy | PHY number (0~4).
@parm uint32* | ptr_linkStatus | PHY link status 1:link up 0:link down
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	Output link status of bit 2 in PHY register 1. API will return 
    status is link up under both auto negotiation complete and link 
    status are set to 1.  
*/
int32 rtl8366rb_getPHYLinkStatus(uint32 phy, uint32 *ptr_linkStatus)
{
	uint32 phyData;

	if(phy > RTL8366RB_PHY_NO_MAX)
		return FAILED;

	/*Get PHY status register*/
	if(SUCCESS != rtl8368s_getAsicPHYRegs(phy,0,RTL8366RB_PHY_STATUS_REG,&phyData))
		return FAILED;

	/*check link status*/
	if(phyData & (1<<2))
	{
		*ptr_linkStatus = 1;
	}
	else
	{
		*ptr_linkStatus = 0;	
	}

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_setCPUPort | Set CPU port with/without inserting CPU tag.
@parm uint32 | port | Port number to be set as CPU port (0~6).
@parm uint32 | noTag | NOT insert Realtek proprietary tag (ethernet length/type 0x8899) to frame 1:not insert 0:insert.
@rvalue ERRNO_API_INPUT | invalid input parameters.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure.
@comm
	The API can set CPU port and enable/disable inserting proprietary CPU tag (Length/Type 0x8899)
	to the frame that transmitting to CPU port. It also can enable/disable forwarding unknown
	destination MAC address frames to CPU port. User can reduce CPU loading by not forwarding
	unknown DA frames to CPU port.
*/
int32 rtl8366rb_setCPUPort(uint32 port, uint32 noTag)
{
	if(port >= RTL8366RB_PORT_MAX)
		return ERRNO_API_INPUT;

	/* clean CPU port first */
	if(rtl8368s_setAsicCpuPortMask(0x00) != SUCCESS)
		return FAILED;	


	if(rtl8368s_setAsicCpuPortMask(0x1<<port) != SUCCESS)
		return FAILED;		

	if(rtl8368s_setAsicCpuDisableInsTag(noTag) != SUCCESS)
		return FAILED;	

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getCPUPort | Get CPU port and its setting.
@parm uint32* | ptr_port | returned CPU port (0~7).
@parm uint32* | ptr_noTag | returned CPU port with insert tag ability 1:not insert 0:insert.
@rvalue ERRNO_API_INPUT | invalid input parameters.
@rvalue ERRNO_API_CPUNOTSET | No CPU port speicifed.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure.
@comm
	The API can get configured CPU port and its setting.
	Return ERRNO_API_CPUNOTSET if CPU port is not speicifed.
*/
int32 rtl8366rb_getCPUPort(uint32 *ptr_port, uint32 *ptr_noTag)
{
	uint32 i;
	uint32 portmsk;	
	
	if(ptr_port == NULL || ptr_noTag == NULL)
		return ERRNO_API_INPUT;

	/* get configured CPU port */
	if(rtl8368s_getAsicCpuPortMask(&portmsk) != SUCCESS)
		return FAILED;	

	if(portmsk)
	{
		/*Return first CPU port only*/
		for(i=0;i<=PORT7;i++)
		{
			if(portmsk & (1<<i))
			{
				*ptr_port = i;
				break;
			}		
		}		
	}		
	else
		return ERRNO_API_CPUNOTSET;
		
	if(rtl8368s_getAsicCpuDisableInsTag(ptr_noTag) != SUCCESS)
		return FAILED;				
			
	return SUCCESS;
}

/*
@func int32 | rtl8366rb_setAgeTimerSpeed | Set LUT agging out speed
@parm uint32 | timer | Agging out timer 0:Has been aged out.
@parm uint32 | speed | Agging out speed 0-fastest 3-slowest.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid parameter.
@comm
 	The API can set LUT agging out period for each entry. Differet {timer, speed} parameter will make asic do agging out task
 	at different period time. Following description is times period for different timer,speed combination setting.
			Timer		1		2		3		4		5		6		7
	Speed
	0					14.3s	28.6s	42.9s	57.2s	1.19m	1.43m	1.67m
	1					28.6s	57.2s	1.43m	1.9m	2.38m	2.86m	3.34m
	2					57.2s	1.9m	2.86m	3.81m	4.77m	5.72m	6.68m
	3					1.9m	3.8m	5.72m	7.63m	9.54m	11.45m	13.36m
 	(s:Second m:Minute)
 */
int32 rtl8366rb_setAgeTimerSpeed( uint32 timer, uint32 speed)
{
    return rtl8368s_setAsicAgeTimerSpeed(timer, speed);
}

/*
@func int32 | rtl8368s_getAgeTimerSpeed | Get LUT agging out speed
@parm uint32* | timer | Agging out timer 0:Has been aged out.
@parm uint32* | speed | Agging out speed 0-fastest 3-slowest.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue ERRNO_INPUT | Invalid parameter.
@comm
 	The API can get LUT agging out period for each entry. 
			Timer		1		2		3		4		5		6		7
	Speed
	0					14.3s	28.6s	42.9s	57.2s	1.19m	1.43m	1.67m
	1					28.6s	57.2s	1.43m	1.9m	2.38m	2.86m	3.34m
	2					57.2s	1.9m	2.86m	3.81m	4.77m	5.72m	6.68m
	3					1.9m	3.8m	5.72m	7.63m	9.54m	11.45m	13.36m
 	(s:Second m:Minute)
 */
int32 rtl8366rb_getAgeTimerSpeed( uint32* timer, uint32* speed)
{
    return rtl8368s_getAsicAgeTimerSpeed(timer, speed);
}

/*
@func int32 | rtl8366rb_initChip | Set chip to default configuration enviroment
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure.
@comm
	The API can set chip registers to default configuration for different release chip model.
*/
int32 rtl8366rb_initChip(void)
{
	uint32 index;
	uint32 ledGroup;
	enum RTL8368S_LEDCONF ledCfg[RTL8366RB_LED_GROUP_MAX];
	uint32 verid;
	
	
	const uint32 chipDefault[][2] = {{0x000B, 0x0001},{0x03A6, 0x0100},{0x3A7, 0x0001},{0x02D1, 0x3FFF},
								{0x02D2, 0x3FFF},{0x02D3, 0x3FFF},{0x02D4, 0x3FFF},{0x02D5, 0x3FFF},
								{0x02D6, 0x3FFF},{0x02D7, 0x3FFF},{0x02D8, 0x3FFF},{0x022B, 0x0688},								
								{0x022C, 0x0FAC},{0x03D0, 0x4688},{0x03D1, 0x01F5},{0x0000, 0x0830},
								{0x02F9, 0x0200},{0x02F7, 0x7FFF},{0x02F8, 0x03FF},{0x0080, 0x03E8},
								{0x0081, 0x00CE},{0x0082, 0x00DA},{0x0083, 0x0230},{0xBE0F, 0x2000},
								{0x0231, 0x422A},{0x0232, 0x422A},{0x0233, 0x422A},{0x0234, 0x422A},
								{0x0235, 0x422A},{0x0236, 0x422A},{0x0237, 0x422A},{0x0238, 0x422A},
								{0x0239, 0x422A},{0x023A, 0x422A},{0x023B, 0x422A},{0x023C, 0x422A},
								{0x023D, 0x422A},{0x023E, 0x422A},{0x023F, 0x422A},{0x0240, 0x422A},
								{0x0241, 0x422A},{0x0242, 0x422A},{0x0243, 0x422A},{0x0244, 0x422A},
								{0x0245, 0x422A},{0x0246, 0x422A},{0x0247, 0x422A},{0x0248, 0x422A},
								{0x0249, 0x0146},{0x024A, 0x0146},{0x024B, 0x0146},{0xBE03, 0xC961},
								{0x024D, 0x0146},{0x024E, 0x0146},{0x024F, 0x0146},{0x0250, 0x0146},
								{0xBE64, 0x0226},{0x0252, 0x0146},{0x0253, 0x0146},{0x024C, 0x0146},
								{0x0251, 0x0146},{0x0254, 0x0146},{0xBE62, 0x3FD0},{0x0084, 0x0320},
								{0x0255, 0x0146},{0x0256, 0x0146},{0x0257, 0x0146},{0x0258, 0x0146},	
								{0x0259, 0x0146},{0x025A, 0x0146},{0x025B, 0x0146},{0x025C, 0x0146},	
								{0x025D, 0x0146},{0x025E, 0x0146},{0x025F, 0x0146},{0x0260, 0x0146},		
								{0x0261, 0xA23F},{0x0262, 0x0294},{0x0263, 0xA23F},{0x0264, 0x0294},	
								{0x0265, 0xA23F},{0x0266, 0x0294},{0x0267, 0xA23F},{0x0268, 0x0294},	
								{0x0269, 0xA23F},{0x026A, 0x0294},{0x026B, 0xA23F},{0x026C, 0x0294},	
								{0x026D, 0xA23F},{0x026E, 0x0294},{0x026F, 0xA23F},{0x0270, 0x0294},	
								{0x02F5, 0x0048},{0xBE09, 0x0E00},{0xBE1E, 0x0FA0},{0xBE14, 0x8448},
								{0xBE15, 0x1007},{0xBE4A, 0xA284},{0xC454, 0x3F0B},{0xC474, 0x3F0B},
								{0xBE48, 0x3672},{0xBE4B, 0x17A7},{0xBE4C, 0x0B15},{0xBE52, 0x0EDD},
								{0xBE49, 0x8C00},{0xBE5B, 0x785C},{0xBE5C, 0x785C},{0xBE5D, 0x785C},
								{0xBE61, 0x368A},{0xBE63, 0x9B84},{0xC456, 0xCC13},{0xC476, 0xCC13},
								{0xBE65, 0x307D},{0xBE6D, 0x0005},{0xBE6E, 0xE120},{0xBE2E, 0x7BAF},		
								{0xFFFF, 0xABCD}};

    const uint32 chipDefault2[][2] = {{0x0000, 0x0830},{0x0400, 0x8130},{0x000A, 0x83ED},{0x0431, 0x5432},
								   {0x0F51, 0x0017},{0x02F5, 0x0048},{0x02FA, 0xFFDF},{0x02FB, 0xFFE0},
                                   {0xC456, 0x0C14},{0xC476, 0x0C14},{0xC454, 0x3F8B},{0xC474, 0x3F8B},
                                   {0xC450, 0x2071},{0xC470, 0x2071},{0xC451, 0x226B},{0xC471, 0x226B},
                                   {0xC452, 0xA293},{0xC472, 0xA293},{0xC44C, 0x1585},{0xC44C, 0x1185},
                                   {0xC44C, 0x1585},{0xC46C, 0x1585},{0xC46C, 0x1185},{0xC46C, 0x1585},
                                   {0xC44C, 0x0185},{0xC44C, 0x0181},{0xC44C, 0x0185},{0xC46C, 0x0185},
                                   {0xC46C, 0x0181},{0xC46C, 0x0185},{0xBE24, 0xB000},{0xBE23, 0xFF51},
                                   {0xBE22, 0xDF20},{0xBE21, 0x0140},{0xBE20, 0x00BB},{0xBE24, 0xB800},
                                   {0xBE24, 0x0000},{0xBE24, 0x7000},{0xBE23, 0xFF51},{0xBE22, 0xDF60},
                                   {0xBE21, 0x0140},{0xBE20, 0x0077},{0xBE24, 0x7800},{0xBE24, 0x0000},
                                   {0xBE2E, 0x7BA7},{0xBE36, 0x1000},{0xBE37, 0x1000},{0x8000, 0x0001},
                                   {0xBE69, 0xD50F},{0x8000, 0x0000},{0xBE69, 0xD50F},{0xBE6B, 0x0320},
                                   {0xBE77, 0x2800},{0xBE78, 0x3C3C},{0xBE79, 0x3C3C},{0xBE6E, 0xE120},
                                   {0x8000, 0x0001},{0xBE10, 0x8140},{0x8000, 0x0000},{0xBE10, 0x8140},
                                   {0xBE15, 0x1007},{0xBE14, 0x0448},{0xBE1E, 0x00A0},{0xBE10, 0x8160},
                                   {0xBE10, 0x8140},{0xBE00, 0x1340},{0x0450, 0x0000},{0x0401, 0x0000},
                                   {0xFFFF, 0xABCD}};
                                   
	for(ledGroup= 0;ledGroup<RTL8366RB_LED_GROUP_MAX;ledGroup++)
	{
		if(SUCCESS != rtl8368s_getAsicLedIndicateInfoConfig(ledGroup,&ledCfg[ledGroup]))
			return FAILED;

		if(SUCCESS != rtl8368s_setAsicLedIndicateInfoConfig(ledGroup,LEDCONF_LEDFORCE))
			return FAILED;
	}

	if(SUCCESS != rtl8368s_setAsicForceLeds(0x3F,0x3F,0x3F,0x3F))
		return FAILED;

	if(SUCCESS != rtl8368s_getAsicReg(RTL8368S_REVISION_ID_REG,&verid))
			return FAILED;
	index = 0;

    switch (verid)
    {
        case 0x0000:
            while(chipDefault[index][0] != 0xFFFF && chipDefault[index][1] != 0xABCD)
    		{	
    			/*PHY registers setting*/	
    			if(0xBE00 == (chipDefault[index][0] & 0xBE00))
    			{
    				if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_WRITE))
    					return FAILED;
    
    			}
    
    			if(SUCCESS != rtl8368s_setAsicReg(chipDefault[index][0],chipDefault[index][1]))
    				return FAILED;
    				
    			index ++;	
    		}
            break;

        case 0x0001:
        case 0x0002:
        case 0x0003:
        default:
            while(chipDefault2[index][0] != 0xFFFF && chipDefault2[index][1] != 0xABCD)
    		{	
    			/*PHY registers setting*/	
    			if(0xBE00 == (chipDefault2[index][0] & 0xBE00))
    			{
    				if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_WRITE))
    					return FAILED;
    
    			}
    
    			if(SUCCESS != rtl8368s_setAsicReg(chipDefault2[index][0],chipDefault2[index][1]))
    				return FAILED;
    				
    			index ++;	
    		}
            break;
    }

	DELAY_800MS_FOR_CHIP_STATABLE();


	for(ledGroup= 0;ledGroup<RTL8366RB_LED_GROUP_MAX;ledGroup++)
	{
		if(SUCCESS != rtl8368s_setAsicLedIndicateInfoConfig(ledGroup,ledCfg[ledGroup]))
			return FAILED;
			
	}

	return SUCCESS;
}



/*
@func int32 | rtl8366rb_setLedMode | Set Led to congiuration mode
@parm rtl8366rb_led_mode_t | mode | Support mode0,mode1,mode2,mode3 and force mode only.
@parm uint32 | ledG0Msk | Turn on or turn off Leds of group 0, 1:on 0:off.
@parm uint32 | ledG1Msk | Turn on or turn off Leds of group 1, 1:on 0:off.
@parm uint32 | ledG2Msk | Turn on or turn off Leds of group 2, 1:on 0:off.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	While led mode was set to FORCE, user can force turn on or turn off dedicted leds of each port. There are three LEDs for each port for indicating information
	about dedicated port. LEDs is set to indicate different information of port in different mode.

	Mode0
		LED0-Link, Activity Indicator. Low for link established. Link/Act Blinks when the corresponding port is transmitting or receiving.
		LED1-1000Mb/s Speed Indicator. Low for 1000Mb/s.
		LED2-100Mb/s Speed Indicator. Low for 100Mb/s.

	Mode1
		LED0-Link, Activity Indicator. Low for link established. Link/Act Blinks when the corresponding port is transmitting or receiving.
		LED1-1000Mb/s Speed Indicator. Low for 1000Mb/s.
		LED2-Collision, Full duplex Indicator. Blinking when collision happens. Low for full duplex, and high for half duplex mode.

	Mode2
		LED0-Collision, Full duplex Indicator. Blinking when collision happens. Low for full duplex, and high for half duplex mode.
		LED1-1000Mb/s Speed/Activity Indicator. Low for 1000Mb/s. Blinks when the corresponding port is transmitting or receiving.
		LED2-10/100Mb/s Speed/Activity Indicator. Low for 10/100Mb/s. Blinks when the corresponding port is transmitting or receiving.

	Mode3
		LED0-10Mb/s Speed/Activity Indicator. Low for 10Mb/s. Blinks when the corresponding port is transmitting or receiving.
		LED1-1000Mb/s Speed/Activity Indicator. Low for 1000Mb/s. Blinks when the corresponding port is transmitting or receiving.
		LED2-100Mb/s Speed/Activity Indicator. Low for 100Mb/s. Blinks every the corresponding port is transmitting or receiving.

	Force: All Leds can be turned on or off by software configuration.
	
	
*/
int32 rtl8366rb_setLedMode(rtl8366rb_led_mode_t mode, rtl8366rb_LedConfig_t *ptr_ledgroup)
{
	uint16 ledGroup;
	const uint32 ledConfig[RTL8366RB_LED_MODE_MAX][RTL8366RB_LED_GROUP_MAX] = { 
					{LEDCONF_LINK_ACT,		LEDCONF_SPD1000,		LEDCONF_SPD100,			LEDCONF_SPD10},
					{LEDCONF_LINK_ACT,		LEDCONF_SPD1000,		LEDCONF_DUPCOL,		LEDCONF_SPD100},
					{LEDCONF_DUPCOL,		LEDCONF_SPD1000ACT,	LEDCONF_SPD10010ACT,	LEDCONF_LEDOFF},
					{LEDCONF_SPD10ACT,		LEDCONF_SPD1000ACT,	LEDCONF_SPD100ACT,		LEDCONF_DUPCOL},
					{LEDCONF_LEDFORCE,		LEDCONF_LEDFORCE,		LEDCONF_LEDFORCE,		LEDCONF_LEDFORCE}
					};
		

	if(mode >= RTL8366RB_LED_MODE_MAX)
		return FAILED;


	for(ledGroup= 0;ledGroup<RTL8366RB_LED_GROUP_MAX;ledGroup++)
	{
		if(SUCCESS != rtl8368s_setAsicLedIndicateInfoConfig(ledGroup,ledConfig[mode][ledGroup]))
			return FAILED;
	}

	if(mode != RTL8366RB_LED_MODE_FORCE)
		return SUCCESS;

	if(SUCCESS != rtl8368s_setAsicForceLeds(ptr_ledgroup->ledG0Msk,ptr_ledgroup->ledG1Msk,ptr_ledgroup->ledG2Msk,0))
		return FAILED;

	return SUCCESS;
}


/*
@func int32 | rtl8366rb_setLedBlinkRate | Set LED blinking rate at mode 0 to mode 3
@parm rtl8366rb_led_blinkrate_t | blinkRate | Support 6 blinking rate.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	ASIC support 6 types of LED blinking rates at 28ms, 56ms, 84ms, 111ms, 222ms and 446ms.
*/
int32 rtl8366rb_setLedBlinkRate(rtl8366rb_led_blinkrate_t blinkRate)
{
	if(blinkRate >= RTL8366RB_LEDBLINKRATE_MAX)
		return FAILED;

	if(SUCCESS != rtl8368s_setAsicLedBlinkRate(blinkRate))
		return FAILED;

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_setRtctTesting | Set port RTCT cable length detection start.
@parm uint32 | operation | Operation mode
@parm enum PORTID | port | Port number to be set as CPU port (0~4).
@rvalue ERRNO_API_INVALIDPARAM | invalid input parameters.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	Detect dedicated port cable length by ASIC with desired channel. 
    After start RTCT detection testing, user must wait more tan 2 
    seconds to get detected result.
*/
int32 rtl8366rb_setRtctTesting(uint32 operation, enum PORTID port)
{
	if(port >= PORT_MAX)
		return ERRNO_INPUT;

    if(operation == 1)
    {
        rtl8368s_setAsicPHYRegs(0, 0, 21, 0x0007);
        rtl8368s_setAsicPHYRegs(1, 0, 21, 0x0007);
        rtl8368s_setAsicPHYRegs(2, 0, 21, 0x0007);
        rtl8368s_setAsicPHYRegs(3, 0, 21, 0x0007);
        rtl8368s_setAsicPHYRegs(4, 0, 21, 0x0007);
    }
		
	if(SUCCESS != rtl8368s_setAsicCableTesting(port, ENABLE))
		return FAILED;

	return SUCCESS;
}
/*
@func int32 | rtl8366rb_getRtctResult | Get port RTCT cable length detection result
@parm uint32 | operation | Operation mode
@parm enum PORTID | port | Port number to be set as CPU port (0~4).
@parm enum CHANNELID | channel | Detected channel id (0~3}.
@parm uint32* |length | Cable length. (unit meter)
@parm uint32* |impmis | Channel impedance mismatch status 1: impedance mismatch, 0: normal
@parm uint32* |openSts | Open status of previous RTCT detection result.
@parm uint32* |shortSts | Short status of previous RTCT detection result.
@rvalue ERRNO_API_INVALIDPARAM | invalid input parameters.
@rvalue ERRNO_API_RTCT_NOT_FINISHED | Still in RTCT detection state.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	After start RTCT detection testing, user must wait more than 
    2 seconds to get detected result. 
*/
int32 rtl8366rb_getRtctResult(uint32 operation, enum PORTID port,enum CHANNELID channel, uint32* length, uint32* impmis, uint32* openSts, uint32* shortSts)
{
    if(port >= PORT_MAX)
		return ERRNO_INPUT;

    if(SUCCESS != rtl8368s_getAsicRtctChannelStatus(port, channel, impmis, openSts, shortSts, length))
		return FAILED;

    if(operation == 1)
    {
        rtl8368s_setAsicPHYRegs(0, 0, 21, 0x1007);
        rtl8368s_setAsicPHYRegs(1, 0, 21, 0x1007);
        rtl8368s_setAsicPHYRegs(2, 0, 21, 0x1007);
        rtl8368s_setAsicPHYRegs(3, 0, 21, 0x1007);
        rtl8368s_setAsicPHYRegs(4, 0, 21, 0x1007);
    }

    return SUCCESS;
}

#if	0
/*
@func int32 | rtl8366rb_setReservedMulticastAddress | Set reserved multicast address frame trap to CPU.
@parm rtl8366rb_rma_frame_t | rma | type of RMA for trapping frame type setting.
@parm enum RMAOP | op | RMA operation 0:forwarding 1:Trap to CPU 2:Drop 3:Forward, exclude CPU.
@parm uint32 | keepCtag | Keep CVLAN tag/untag format 1:enable 0:disable.
@parm uint32 | bypassStorm | By pass storm control calculation 1:enable 0:disable.
@parm uint32 | priSel | New CPU priority assignment 1:enable 0:disable.
@parm enum PRIORITYVALUE | priority | Priority for new CPU priority assignment. This CPU priority will not be remapped.
@rvalue ERRNO_INPUT | Invalid input parameter.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	There are 23 types of Reserved Multicast Address frame for trapping to CPU for application usage. They are as following definition.

	DMAC						Assignment
	01-80-C2-00-00-00			Bridge Group Address
	01-80-C2-00-00-01			IEEE Std 802.3, 1988 Edition, Full Duplex PAUSE operation
	01-80-C2-00-00-02			IEEE Std 802.3ad Slow Protocols-Multicast address
	01-80-C2-00-00-03			IEEE Std 802.1X PAE address
	01-80-C2-00-00-08			Provider Bridge Group Address
	01-80-C2-00-00-0D			Provider Bridge GVRP Address
	01-80-C2-00-00-0E			IEEE Std 802.1ab Link Layer Discovery Protocol Multicast address
	01-80-C2-00-00-10			All LANs Bridge Management Group Address
	01-80-C2-00-00-20			GMRP Address
	01-80-C2-00-00-21			GVRP address
	01-80-C2-00-00-04,05,06,07,	Undefined 802.1 bridge address
				09,0A,0B,0C,0F		
	01-80-C2-00-00-22~2F		Undefined GARP address
	xx-xx-xx-xx-xx-xx				IGMPin PPPoE frame(Type/Length = 8864)
	xx-xx-xx-xx-xx-xx				MLD(IPv6 ICMP) in PPPoE frame(Type/Length = 8864)
	xx-xx-xx-xx-xx-xx				IGMP packet but not PPPoE frame
	xx-xx-xx-xx-xx-xx				MLD(IPv6 ICMP) but not PPPoE frame
	xx-xx-xx-xx-xx-xx				User defined address 1
	xx-xx-xx-xx-xx-xx				User defined address 2
	xx-xx-xx-xx-xx-xx				User defined address 3
	xx-xx-xx-xx-xx-xx				Unknow source address


	
*/
int32 rtl8366rb_setReservedMulticastAddress(rtl8366rb_rma_frame_t rma, rtl8366rb_rmaConfig_t *ptr_rmacfg)
{
	if(rma> RMA_MAX)
		return ERROR_RMA_INDEX;

	if(ptr_rmacfg->op >=RMAOP_MAX)
		return ERRNO_INPUT;
	
	if(ptr_rmacfg->priority >=PRI_MAX)
		return ERRNO_PRIORITY;

	if(SUCCESS != rtl8368s_setAsicRma(rma,ptr_rmacfg->op,ptr_rmacfg->keepCtag,ptr_rmacfg->bypassStorm,ptr_rmacfg->priSel,ptr_rmacfg->priority))
		return FAILED;		

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getReservedMulticastAddress | Set reserved multicast address frame trap to CPU.
@parm rtl8366rb_rma_frame_t | rma | type of RMA for trapping frame type setting.
@parm enum RMAOP* | op | RMA operation 0:forwarding 1:Trap to CPU 2:Drop 3:Forward, exclude CPU.
@parm uint32* | keepCtag | Keep CVLAN tag/untag format 1:enable 0:disable.
@parm uint32* | bypassStorm | By pass storm control calculation 1:enable 0:disable.
@parm uint32* | priSel | New CPU priority assignment 1:enable 0:disable.
@parm enum* PRIORITYVALUE | priority | Priority for new CPU priority assignment. This CPU priority will not be remapped.
@rvalue ERRNO_INPUT | Invalid input parameter.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	The API can retrieved RMA configuration. 
*/
int32 rtl8366rb_getReservedMulticastAddress(rtl8366rb_rma_frame_t rma, rtl8366rb_rmaConfig_t *ptr_rmacfg)
{
	if(rma> RMA_MAX)
		return ERRNO_INPUT;


	if(SUCCESS != rtl8368s_getAsicRma(rma,&ptr_rmacfg->op,&ptr_rmacfg->keepCtag,&ptr_rmacfg->bypassStorm,&ptr_rmacfg->priSel,&ptr_rmacfg->priority))
		return FAILED;		

	return SUCCESS;
}


/*
@func int32 | rtl8366rb_setPortMirror | Configure port mirror function.
@parm enum RTL8366RB_PORTID | rtl8366rb_portMirror_t->sourcePort | Source port.
@parm enum RTL8366RB_PORTID | rtl8366rb_portMirror_t->monitorPort | Monitor port.
@parm uint32 | rtl8366rb_portMirror_t->mirRx | Rx mirror function 1:enable 0:disable.
@parm uint32 | rtl8366rb_portMirror_t->mirTx | Tx mirror function 1:enable 0:disable. 
@parm uint32 | rtl8366rb_portMirror_t->mirIso | Mirror isolation for monitor port 1:enable 0:disable. 
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	ASIC supports one port mirror function only. For mirror function, there are one mirrored port and one monitor port. Frame from/to source port
	can be mirrored to monitor port and mirror port can be set with receiving frame from source port only. Mirror configuration is as below description.
	Field					Description
	sourcePort			Desired source port which receiving/transmitting frame to be mirrored 
	monitorPort			Monitor port which frames were mirrored to.
	mirRx				Function setting of mirroring frames received in source port.
	mirTx				Funciton setting of mirroring frames transmitted from source port.
	mirIso				Setting which permit forwarding frame from source port only.
*/
int32 rtl8366rb_setPortMirror(rtl8366rb_portMirror_t *ptr_mirrorcfg)
{
	if(SUCCESS != rtl8368s_setAsicPortMirror(ptr_mirrorcfg->sourcePort,ptr_mirrorcfg->monitorPort))
		return FAILED;		

	if(SUCCESS != rtl8368s_setAsicPortMirrorRxFunction(ptr_mirrorcfg->mirrorRx))
		return FAILED;		

	if(SUCCESS != rtl8368s_setAsicPortMirrorTxFunction(ptr_mirrorcfg->mirrorTx))
		return FAILED;		

	if(SUCCESS != rtl8368s_setAsicPortMirrorIsolation(ptr_mirrorcfg->mirrorIso))
		return FAILED;		

	return SUCCESS;

}
#endif

/*
@func int32 | rtl8366rb_setPort4RGMIIControl | Set LED/Interrupt Control for P4 RGMII mode
@parm uint32 | enabled | 1:Enable LED/Interrupt for P4 RGMII, 0:Disable LED/Interrupt for P4 RGMII
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@comm
	Set LED/Interrupt Control for P4 RGMII mode
*/
int32 rtl8366rb_setPort4RGMIIControl(uint32 enabled)
{
    return rtl8368s_setAiscPort4RGMIIControl(enabled);
}

/*
@func int32 | rtl8366rb_setMac5ForceLink | Set Port/MAC 5 force linking configuration.
@parm enum PORTLINKSPEED | speed | Speed of the port 0b00-10M, 0b01-100M,0b10-1000M, 0b11 reserved.
@parm enum PORTLINKDUPLEXMODE | duplex | Deuplex mode 0b0-half duplex 0b1-full duplex.
@parm uint32 | link | Link status 0b0-link down b1-link up.
@parm uint32 | txPause | Pause frame transmit ability of the port 0b0-inactive 0b1-active.
@parm uint32 | rxPause | Pause frame response ability of the port 0b0-inactive 0b1-active.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
  	This API can set Port/MAC 5 properties in force mode. 
 */
int32 rtl8366rb_setMac5ForceLink(rtl8366rb_macConfig_t *ptr_maccfg)
{
	if(ptr_maccfg->speed > SPD_1000M)
		return ERROR_MAC_INVALIDSPEED;

	if(SUCCESS != rtl8368s_setAsicMacForceLink(PORT5,1,ptr_maccfg->speed,ptr_maccfg->duplex,ptr_maccfg->link,ptr_maccfg->txPause,ptr_maccfg->rxPause))
		return FAILED;

	return SUCCESS;
}
/*
@func int32 | rtl8366rb_getMac5ForceLink | Set Port/MAC 5 force linking configuration.
@parm enum PORTLINKSPEED* | speed | Speed of the port 0b00-10M, 0b01-100M,0b10-1000M, 0b11 reserved.
@parm enum PORTLINKDUPLEXMODE* | duplex | Deuplex mode 0b0-half duplex 0b1-full duplex.
@parm uint32* | link | Link status 0b0-link down b1-link up.
@parm uint32* | txPause | Pause frame transmit ability of the port 0b0-inactive 0b1-active.
@parm uint32* | rxPause | Pause frame response ability of the port 0b0-inactive 0b1-active.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
  	This API can set Port/MAC 5 properties in force mode. 
 */
int32 rtl8366rb_getMac5ForceLink(rtl8366rb_macConfig_t *ptr_maccfg)
{
	uint32 autoNegotiation;
	if(SUCCESS != rtl8368s_getAsicPortLinkState(PORT5,&ptr_maccfg->speed,&ptr_maccfg->duplex,&ptr_maccfg->link,&ptr_maccfg->txPause,&ptr_maccfg->rxPause,&autoNegotiation))
		return FAILED;

	return SUCCESS;
}



/*
@func int32 | rtl8366rb_setStormFiltering | Set per port  storm filtering function.
@parm rtl8366rb_stormfilter_t* | ptr_stormftr | The structure of Sotrm filtering function.
@rvalue ERRNO_API_INPUT | Input NULL pointer.
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	The API can be used to enable or disable per-port storm filtering 
    control. There are three kinds of storm filtering control which 
    are broadcasting storm control, multicast storm control and unknown 
    unicast storm control. There are dedicated storm control setting 
    for each port and global storm control meter for each one reference. 
    There is still a didicated reference meter counter in each port.
*/
int32 rtl8366rb_setStormFiltering(rtl8366rb_stormfilter_t *ptr_stormftr)
{
    if(ptr_stormftr == NULL)
		return ERRNO_API_INPUT;
	
	if(ptr_stormftr->port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	if(SUCCESS != rtl8368s_setAsicStormFiltering(ptr_stormftr->port, ptr_stormftr->bcstorm, ptr_stormftr->mcstorm, ptr_stormftr->undastorm, ptr_stormftr->unmcstorm))
		return FAILED;

	if(SUCCESS != rtl8368s_setAsicStormFilteringMeter(ptr_stormftr->port, ptr_stormftr->rate, ptr_stormftr->ifg))
		return FAILED;

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getStormFiltering | Get per port storm filtering function.
@parm rtl8366rb_stormfilter_t* | ptr_stormftr | The structure of Sotrm filtering function.
@rvalue ERRNO_API_INPUT | Input NULL pointer.
@rvalue ERRNO_PORT_NUM | Invalid Port Number.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	The API can be used to get per-port storm filtering control. 
    There are four kinds of storm filtering control which are 
    broadcasting storm control, multicast storm control unknown 
    unicast storm control and unknown multicast storm control. 
    There are dedicated storm control setting for each port 
    and global storm control meter for each one reference. 
    There is still a didicated reference meter counter in each port.
*/
int32 rtl8366rb_getStormFiltering(rtl8366rb_stormfilter_t *ptr_stormftr)
{
    if(ptr_stormftr == NULL)
        return ERRNO_API_INPUT;

	if(ptr_stormftr->port >=PORT_MAX)
		return ERRNO_PORT_NUM;

	if(SUCCESS != rtl8368s_getAsicStormFiltering(ptr_stormftr->port,
                                                 &(ptr_stormftr->bcstorm),
                                                 &(ptr_stormftr->mcstorm),
                                                 &(ptr_stormftr->undastorm),
                                                 &(ptr_stormftr->unmcstorm)))
		return FAILED;

	if(SUCCESS != rtl8368s_getAsicStormFilteringMeter(ptr_stormftr->port,
                                                      &(ptr_stormftr->rate),
                                                      &(ptr_stormftr->ifg)))
		return FAILED;
		
	return SUCCESS;
}


/*
@func int32 | rtl8366rb_setPortIsolation | Set port isolation function  usage configuration.
@parm enum RTL8366RB_PORTID | ptr_isocfg->port | Egress port number (0~8).
@parm uint32 | ptr_isocfg->enable | Isolation function usage 1:enable, 0:disabled.
@parm uint32 | ptr_isocfg->portmask | Output port mask (0-8).
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMI | SMI access error.
@rvalue FAILED | Invalid parameter.
@comm
	The API can isolation source port output to output ports 
    represented by port mask. It is enabled per port. When 
	the API is used, packets from source port will not deliver 
    to other ports that are not in port mask. Other ports not 
    in port mask can still deliver packets to source port. 
    By using isolation setting, user can isolate physical ports' 
    group without VLAN configuration. This function will physically 
    seperate networking flow and isolate network within desired ports only.
 */

int32 rtl8366rb_setPortIsolation(rtl8366rb_isolation_t *ptr_isocfg)
{

	if(ptr_isocfg->port >=RTL8366RB_PORT_MAX)
		return ERRNO_INPUT;

	if(rtl8368s_setAsicPortIsolation(ptr_isocfg->port, ptr_isocfg->enable)!=SUCCESS)
		return FAILED;

	if(rtl8368s_setAsicPortIsolationConfig(ptr_isocfg->port, ptr_isocfg->portmask)!=SUCCESS)
		return FAILED;


	return SUCCESS;
}
/*
@func int32 | rtl8366rb_getPortIsolation | Get ASIC port isolation control configuration. 
@parm enum RTL8366RB_PORTID | ptr_isocfg->port | Source port number.
@parm uint32 | ptr_isocfg->enable | Isolation function usage
@parm uint32 | ptr_isocfg-> portmask | Output port mask.
@rvalue SUCCESS | Success.
@rvalue ERRNO_SMIERROR | SMI access error.
@rvalue ERRNO_INPUT | Invalid port number.
@comm
 	The API can be used to get the port isolation configuration.
*/

int32 rtl8366rb_getPortIsolation(rtl8366rb_isolation_t *ptr_isocfg)
{

	if(ptr_isocfg->port >=RTL8366RB_PORT_MAX)
		return ERRNO_INPUT;

	if(rtl8368s_getAsicPortIsolation(ptr_isocfg->port, &ptr_isocfg->enable)!=SUCCESS)
		return FAILED;

	if(rtl8368s_getAsicPortIsolationConfig(ptr_isocfg->port, &ptr_isocfg->portmask)!=SUCCESS)
		return FAILED;


	return SUCCESS;
}

#if	0
/*
@func int32 | rtl8366rb_setSourceMacLearning | Set per-port source mac learning function.
@parm uint32 | port | Physical port number (0~5).
@parm uint32 | enabled | Source mac learning function 1:enable 0:disable.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	Like CPU port usage, the source mac of frame from CPU port 
    will be learned if this frame is just traped to CPU port 
    with frame modification only and re-send from CPU. This frame 
    will be learned twiced and the source port information of 
    L2 entry about this frame maybe is wrong. 
*/
int32 rtl8366rb_setSourceMacLearning(uint32 port, uint32 enabled)
{
    if(port >= RTL8366RB_PORT_MAX)
		return FAILED;

    if(SUCCESS != rtl8368s_setAsicPortLearnDisable(port, enabled?FALSE:TRUE))
		return FAILED;

	return SUCCESS;
}

/*
@func int32 | rtl8366rb_getSourceMacLearning | Get per-port source mac learning function.
@parm uint32 | port | Physical port number (0~5).
@parm uint32* | ptr_enabled | Source mac learning function 1:enable 0:disable.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure. 
@comm
	Like CPU port usage, the source mac of frame from CPU port 
    will be learned if this frame is just traped to CPU port 
    with frame modification only and re-send from CPU. This frame 
    will be learned twiced and the source port information of L2 
    entry about this frame maybe is wrong. 
*/
int32 rtl8366rb_getSourceMacLearning(uint32 port, uint32 *ptr_enabled)
{
    uint32  disabled;

	if(port >= RTL8366RB_PORT_MAX)
		return FAILED;

    if(SUCCESS != rtl8368s_getAsicPortLearnDisable(port, &disabled))
		return FAILED;

    *ptr_enabled = (disabled ? 0 : 1);
    return SUCCESS;
}
#endif

/*
@func int32 | rtl8366rb_setGreenEthernet | Set green ethernet function.
@parm uint32 | greenFeature | Green feature function usage 1:enable 0:disable.
@parm uint32 | powerSaving | Power saving mode 1:power saving mode 0:normal mode.
@rvalue SUCCESS | Success.
@rvalue FAILED | Failure.
@comm
 	The API can set Green Ethernet function to reduce power consumption. 
    While green feature is enabled, ASIC will automatic detect the cable 
    length and then select different power mode for best performance with 
    minimums power consumption. Link down ports will enter power savining 
    mode in 10 seconds after the cable disconnected if power saving function 
    is enabled.
*/
int32 rtl8366rb_setGreenEthernet(uint32 greenFeature, uint32 powerSaving)
{
	uint32 phyNo;
	uint32 regData;
	uint32 idx;
	const uint32 greenSeting[][2] = {	{0xBE78,0x323C},	{0xBE77,0x5000}, {0xBE2E,0x7BA7},
									{0xBE59,0x3459},	{0xBE5A,0x745A},	{0xBE5B,0x785C},
									{0xBE5C,0x785C}, {0xBE6E,0xE120},	{0xBE79,0x323C},	
									{0xFFFF,0xABCD}
								};

	idx = 0;	
	while(greenSeting[idx][0] != 0xFFFF && greenSeting[idx][1] != 0xABCD)
	{
		if(SUCCESS != rtl8368s_getAsicReg(RTL8368S_PHY_ACCESS_BUSY_REG, &regData))
				return FAILED;
		if((regData&RTL8368S_PHY_INT_BUSY_MASK)==0)
		{
			/*PHY registers setting*/	
			if(SUCCESS != rtl8368s_setAsicReg(RTL8368S_PHY_ACCESS_CTRL_REG, RTL8368S_PHY_CTRL_WRITE))
				return FAILED;
			if(SUCCESS != rtl8368s_setAsicReg(greenSeting[idx][0],greenSeting[idx][1]))
				return FAILED;
			idx ++;
		}
	}	
		
	if(SUCCESS !=rtl8368s_setAsicGreenFeature(greenFeature))
		return FAILED;

	for(phyNo = 0; phyNo<=RTL8366RB_PHY_NO_MAX;phyNo++)
	{
		if(SUCCESS !=rtl8368s_setAsicPowerSaving(phyNo,powerSaving))
			return FAILED;	
	}

	return SUCCESS;
}






//#######################################################################
//	Add Buffalo
//#######################################################################
/******************************************************************************
*
* rtl8366rb_phy_is_link_alive - test to see if the specified link is alive
*
* RETURNS:
*    TRUE  --> link is alive
*    FALSE --> link is down
*/

BOOL rtl8366rb_phy_is_link_alive(int phyUnit)
{
    uint32_t regData;

	if (SUCCESS==rtl8366rb_getPHYLinkStatus(phyUnit,&regData)) {
		if (regData)
			return	TRUE;
	}

	return FALSE;
}

/******************************************************************************
*
* rtl8366rb_phy_setup - reset and setup the PHY associated with
* the specified MAC unit number.
*
* Resets the associated PHY port.
*
* RETURNS:
*    TRUE  --> associated PHY is alive
*    FALSE --> no LINKs on this ethernet unit
*/

BOOL
rtl8366rb_phy_setup(int ethUnit)
{
	int phyUnit,phyStart,phyEnd;
	int liveLinks = 0;    
	uint32_t regData=0;
	enum PORTLINKSPEED speed=0;
	enum PORTLINKDUPLEXMODE duplex=0;
	uint32_t link=0,txPause=0,rxPause=0,autoNegotiation=0; 

	static	int init_first=0;

	printf("%s  ethUnit=%x\n", __FUNCTION__, ethUnit);    

	if (!init_first) {
		uint32_t	regID, regREV;

//		smi_init();

		rtl8368s_getAsicReg(RTL8368S_CHIP_ID_REG, &regID);
		rtl8368s_getAsicReg(RTL8368S_REVISION_ID_REG, &regREV);
		printf("Realtek 8366RB switch ID:0x%x REV:0x%x\n", regID, regREV);

		//	initialize chip
		rtl8366rb_initChip();   

		//	wait initialize
		sysMsDelay(4000);    
	
#ifdef	CPU_PORT	
		/* Set port 5 noTag and don't dropUnda */	
		rtl8366rb_setCPUPort(PORT5, 1, 0);	
#ifdef GET_INFO
		rtl8366rb_getCPUPort(&speed,&duplex,&link);
#endif	//GET_INFO
#endif	//CPU_PORT



		/* Enable CPU port - forced speed, full duplex and flow control */
#ifdef	CFG_SP1000
		{
			rtl8366rb_macConfig_t mac5_cfg;
			mac5_cfg.force = MAC_FORCE;
			mac5_cfg.speed = SPD_1000M;
			mac5_cfg.duplex = FULL_DUPLEX;
			mac5_cfg.link = 1;
			mac5_cfg.txPause = 1;
			mac5_cfg.rxPause = 1;
			DEBUG_MSG("SP1000\n");
			rtl8366rb_setMac5ForceLink(&mac5_cfg);
#ifdef	GET_INFO
			rtl8366rb_getMac5ForceLink(&mac5_cfg);
#endif	//GET_INFO
		}
#else //100
		{
			rtl8366rb_macConfig_t mac5_cfg;
			mac5_cfg.force = MAC_FORCE;
			mac5_cfg.speed = SPD_100M;
			mac5_cfg.duplex = FULL_DUPLEX;
			mac5_cfg.link = 1;
			mac5_cfg.txPause = 1;
			mac5_cfg.rxPause = 1;
			DEBUG_MSG("SP1000\n");
			rtl8366rb_setMac5ForceLink(&mac5_cfg);
#ifdef	GET_INFO
			rtl8366rb_getMac5ForceLink(&mac5_cfg);
#endif	//GET_INFO
		}
#endif	//SPEED

		rtl8368s_setAsicReg(0xC44C, 0x0085);

		//	Enable GreenEthernet
		if (SUCCESS != rtl8366rb_setGreenEthernet(1, 1)) {
			printf("Realtek Greeen Ethernet Setup Fail!\n");
		} else {
			printf("Realtek Greeen Ethernet Setup\n");
		}


		init_first = 1;
	}

	if(ethUnit){//wan
		phyStart = 4;
		phyEnd = REALTEK_MAX_PORT_ID;
	}
	else{//lan
		phyStart = 0;
		phyEnd = REALTEK_MAX_PORT_ID-1;	 
	}

	for (phyUnit=phyStart; phyUnit < phyEnd; phyUnit++) {
#if 0//def GET_INFO		                
        rtl8368s_getAsicPortLinkState(phyUnit,&speed,&duplex,&link,&txPause,&rxPause,&autoNegotiation);
        DEBUG_MSG("\nSW port[%d] init state:", phyUnit);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_CONTROL_REG,&regData);
        DEBUG_MSG("\nBasic Control: REG[0] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_STATUS_REG,&regData);
        DEBUG_MSG("\nBasic Status:  REG[1] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_PHY_ID0,&regData);
        DEBUG_MSG("\nPHY ID  :      REG[2] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_PHY_ID1,&regData);
        DEBUG_MSG("\n               REG[3] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_LOCAL_CAP,&regData);
        DEBUG_MSG("\nAutoNeg Local: REG[4] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_REMOTE_CAP,&regData);
        DEBUG_MSG("\n       Remote: REG[5] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_GIGA_CONTROL,&regData);
        DEBUG_MSG("\nGiga Control:  REG[9] = 0x%hx",regData);
        rtl8368s_getAsicPHYRegs(phyUnit,0,MII_GIGA_STATUS,&regData);
        DEBUG_MSG("\n     Status:   REG[A] = 0x%hx\n",regData);   
#endif
		if (rtl8366rb_phy_is_link_alive(phyUnit)) {
        	DEBUG_MSG("phyUnit=%x is link\n",phyUnit);
            liveLinks++;  
        }  
    }  
#if 0//GET_INFO
    switch_reg_read(0x60,&regData);
    DEBUG_MSG("0x60=0x%x\n", regData);
    switch_reg_read(0x61,&regData);
    DEBUG_MSG("0x61=0x%x\n", regData);
    switch_reg_read(0x62,&regData);
    DEBUG_MSG("0x62=0x%x\n", regData);
	
	switch_reg_read(RTL8368S_CHIP_ID_REG,&regData);	
	DEBUG_MSG("Realtek 8366RB switch ID 0x%x\n", regData);//0x8366
#endif
    return (liveLinks > 0);
}


/****************************************************************************/
BOOL rtl8366rb_phy_is_up(int ethUnit)
{
	uint32_t regData;
	int phyUnit,phyStart,phyEnd;
	int liveLinks = 0; 

	DEBUG_MSG("%s  %d\n",__FUNCTION__, ethUnit);
#if 1
	if(ethUnit){//wan
		phyStart = 4;
		phyEnd = REALTEK_MAX_PORT_ID;
	}
	else{//lan
		phyStart = 0;
		phyEnd = REALTEK_MAX_PORT_ID-1;	 
	}

	for (phyUnit=phyStart; phyUnit < phyEnd; phyUnit++) {
		if (rtl8366rb_phy_is_link_alive(phyUnit)) {
			DEBUG_MSG("phyUnit=%x is link\n",phyUnit);
			liveLinks++;  
		}
	}
	return (liveLinks > 0);	
#else
	return TRUE;
#endif
}


/******************************************************************************
*
* rtl8366rb_phy_is_fdx - Determines whether the phy ports associated with the
* specified device are FULL or HALF duplex.
*
* RETURNS:
*    1  --> FULL
*    0 --> HALF
*/
int
rtl8366rb_phy_is_fdx(int ethUnit)
{
	int	phyUnit,phyStart,phyEnd;  	

	DEBUG_MSG("%s  %d\n",__FUNCTION__, ethUnit);
#if 1  	
	if(ethUnit){//wan
		phyStart = 4;
		phyEnd = REALTEK_MAX_PORT_ID;
		DEBUG_MSG("wan\n");
	}
	else{//lan
		phyStart = 0;
		phyEnd = REALTEK_MAX_PORT_ID-1;	
		DEBUG_MSG("lan\n");
		return TRUE;
	}
#ifdef GET_INFO    	
	for (phyUnit=phyStart; phyUnit < phyEnd; phyUnit++) {       
		if (rtl8366rb_phy_is_link_alive(phyUnit)) {
			DEBUG_MSG("phyUnit=%x is link\n",phyUnit);
			uint32_t link_status_giga;
			uint32_t local_cap,remote_cap,common_cap;
			
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_GIGA_STATUS,&link_status_giga);
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_LOCAL_CAP,&local_cap);
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_REMOTE_CAP,&remote_cap);
        	common_cap = (local_cap & remote_cap);
        	
        	if (link_status_giga & (1 << MII_GIGA_STATUS_FULL)) {//
            	return TRUE;
            	//link_speed = 1000;
        	} else if (link_status_giga & (1 << MII_GIGA_STATUS_HALF)) {
            	return FALSE;
            	//link_speed = 1000;
        	} else {
            	if (common_cap & (1 << MII_CAP_100BASE_TX_FULL)) {
                	return TRUE;
                	//link_speed = 100;
            	} else if (common_cap & (1 << MII_CAP_100BASE_TX)) {
                	return FALSE;
                	//link_speed = 100;
            	} else if (common_cap & (1 << MII_CAP_10BASE_TX_FULL)) {
                	return TRUE;
                	//link_speed = 10;
           	 	} else  {
                	return FALSE;
                	//link_speed = 10;
           		}
        	}
           
        }
        else{
        	DEBUG_MSG("phyUnit=%x is no link\n",phyUnit);
        }	
    }
#endif    
    return FALSE;
#else
	return TRUE;
#endif
}


/******************************************************************************
*
* rtl8366rb_phy_speed - Determines the speed of phy ports associated with the
* specified device.
*
* RETURNS:
*               AG7100_PHY_SPEED_10T, AG7100_PHY_SPEED_100TX;
*               AG7100_PHY_SPEED_1000T;
*/

int
rtl8366rb_phy_speed(int ethUnit)
{
   int	phyUnit,phyStart,phyEnd;
   
   DEBUG_MSG("%s  %d\n", __FUNCTION__, ethUnit);
#if 1  	
  	if(ethUnit){//wan
    	phyStart = 4;
    	phyEnd = REALTEK_MAX_PORT_ID;
    	DEBUG_MSG("wan\n");
    }	
    else{//lan
    	phyStart = 0;
    	phyEnd = REALTEK_MAX_PORT_ID-1;	
#ifdef CFG_SP1000
		DEBUG_MSG("lan SP1000\n");
		return _1000BASET;
#else
		DEBUG_MSG("lan SP100\n");
		return _100BASET;
#endif	    	 
    }
    	 	
#ifdef GET_INFO   	
    for (phyUnit=phyStart; phyUnit < phyEnd; phyUnit++) {       
        if (rtl8366rb_phy_is_link_alive(phyUnit)) {
        	DEBUG_MSG("phyUnit=%x is link\n",phyUnit);
			uint32_t link_status_giga;
			uint32_t local_cap,remote_cap,common_cap;
			
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_GIGA_STATUS,&link_status_giga);
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_LOCAL_CAP,&local_cap);
			rtl8368s_getAsicPHYRegs(phyUnit,0,MII_REMOTE_CAP,&remote_cap);
        	common_cap = (local_cap & remote_cap);
        	
        	if (link_status_giga & (1 << MII_GIGA_STATUS_FULL)) {//
            	//full_duplex = TRUE;
            	return _1000BASET;
        	} else if (link_status_giga & (1 << MII_GIGA_STATUS_HALF)) {
            	//full_duplex = FALSE;
            	return _1000BASET;
        	} else {
            	if (common_cap & (1 << MII_CAP_100BASE_TX_FULL)) {
                	//full_duplex = TRUE;
                	return _100BASET;
            	} else if (common_cap & (1 << MII_CAP_100BASE_TX)) {
                	//full_duplex = FALSE;
                	return _100BASET;
            	} else if (common_cap & (1 << MII_CAP_10BASE_TX_FULL)) {
                	//full_duplex = TRUE;
                	return _10BASET;
           	 	} else  {
                	//full_duplex = FALSE;
                	return _10BASET;
           		}
        	}
           
        }
        else{
        	DEBUG_MSG("phyUnit=%x is no link\n",phyUnit);
        }	
    }    
#endif    
    return _10BASET;
#else
#ifdef CFG_SP1000
	return _1000BASET;
#else
	return _100BASET;
#endif	
#endif    
}

int	rtl8366rb_set_funcs(void)
{
	g_rtl8366_funcs.pfn_phy_is_up		= rtl8366rb_phy_is_up;
	g_rtl8366_funcs.pfn_phy_is_fdx		= rtl8366rb_phy_is_fdx;
	g_rtl8366_funcs.pfn_phy_speed		= rtl8366rb_phy_speed;
	g_rtl8366_funcs.pfn_phy_setup		= rtl8366rb_phy_setup;
}
